# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Notebooklet for Network Flow Summary."""
from ipaddress import ip_address
from itertools import chain
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple

import pandas as pd
from bokeh.models import LayoutDOM
from IPython.display import display
from msticpy.common.timespan import TimeSpan
from msticpy.datamodel import entities

try:
    from msticpy import nbwidgets
    from msticpy.context.ip_utils import get_ip_type, get_whois_df
    from msticpy.context.ip_utils import ip_whois as get_whois_info
    from msticpy.context.tiproviders.ti_provider_base import ResultSeverity
    from msticpy.vis import foliummap
    from msticpy.vis.timeline import display_timeline, display_timeline_values
except ImportError:
    # Fall back to msticpy locations prior to v2.0.0
    from msticpy.nbtools import foliummap, nbwidgets
    from msticpy.nbtools.nbdisplay import display_timeline, display_timeline_values
    from msticpy.sectools.ip_utils import get_ip_type, get_whois_df, get_whois_info
    from msticpy.sectools.tiproviders.ti_provider_base import ResultSeverity

from .... import nb_metadata
from ...._version import VERSION
from ....common import (
    MsticnbMissingParameterError,
    nb_data_wait,
    nb_markdown,
    nb_warn,
    set_text,
)
from ....data_providers import DataProviders
from ....nblib.azsent.host import get_aznet_topology, get_heartbeat
from ....notebooklet import NBMetadata, Notebooklet, NotebookletResult

__version__ = VERSION
__author__ = "Ian Hellen"


_CLS_METADATA: NBMetadata
_CELL_DOCS: Dict[str, Any]
_CLS_METADATA, _CELL_DOCS = nb_metadata.read_mod_metadata(__file__, __name__)


# pylint: disable=too-few-public-methods, too-many-instance-attributes
class NetworkFlowResult(NotebookletResult):
    """
    Network Flow Details Results.

    Attributes
    ----------
    host_entity : msticpy.datamodel.entities.Host
        The host entity object contains data about the host
        such as name, environment, operating system version,
        IP addresses and Azure VM details. Depending on the
        type of host, not all of this data may be populated.
    network_flows : pd.DataFrame
        The raw network flows recorded for this host.
    plot_flows_by_protocol : LayoutDOM
        Bokeh timeline plot of flow events by protocol.
    plot_flows_by_direction : LayoutDOM
        Bokeh timeline plot of flow events by direction (in/out).
    plot_flow_values : LayoutDOM
        Bokeh values plot of flow events by protocol.
    flow_index : pd.DataFrame
        Summarized DataFrame of flows
    flow_index_data : pd.DataFrame
        Raw summary data of flows.
    flow_summary : pd.DataFrame
        Summarized flows grouped by ASN
    ti_results : pd.DataFrame
        Threat Intelligence results for selected IP Addreses.
    geo_map : foliummap.FoliumMap
        Folium map showing locations of all IP Addresses.
    geo_map_selected : foliummap.FoliumMap
        Folium map showing locations of selected IP Addresses.

    """

    def __init__(
        self,
        description: Optional[str] = None,
        timespan: Optional[TimeSpan] = None,
        notebooklet: Optional["Notebooklet"] = None,
    ):
        """
        Create new Notebooklet result instance.

        Parameters
        ----------
        description : Optional[str], optional
            Result description, by default None
        timespan : Optional[TimeSpan], optional
            TimeSpan for the results, by default None
        notebooklet : Optional[, optional
            Originating notebooklet, by default None

        """
        super().__init__(description, timespan, notebooklet)
        self.description: str = "Network flow results"
        self.host_entity: entities.Host = None
        self.network_flows: Optional[pd.DataFrame] = None
        self.plot_flows_by_protocol: Optional[LayoutDOM] = None
        self.plot_flows_by_direction: Optional[LayoutDOM] = None
        self.plot_flow_values: Optional[LayoutDOM] = None
        self.flow_index: Optional[pd.DataFrame] = None
        self.flow_index_data: Optional[pd.DataFrame] = None
        self.flow_summary: Optional[pd.DataFrame] = None
        self.ti_results: Optional[pd.DataFrame] = None
        self.geo_map: foliummap.FoliumMap = None


class NetworkFlowSummary(Notebooklet):
    """
    Network Flow Summary Notebooklet class.

    Queries network data and plots time lines for network
    traffic to/from a host or IP address.

    - Plot flows events by protocol and direction
    - Plot flow count by protocol
    - Display flow summary table
    - Display flow summary by ASN
    - Display results on map

    Methods
    -------
    - run: main method for notebooklet.
    - select_asns: Open an interactive dialog to choose which ASNs to
      investigate further.
    - lookup_ti_for_asn_ips: For selected ASNs, lookup Threat Intelligence
      data for the IPs belonging to those ASNs.
    - show_selected_asn_map: Show IP address locations for selected IP
      (including any threats highlighted)

    """

    # assign metadata from YAML to class variable
    metadata = _CLS_METADATA
    __doc__ = nb_metadata.update_class_doc(__doc__, metadata)
    _cell_docs = _CELL_DOCS

    def __init__(self, data_providers: Optional[DataProviders] = None, **kwargs):
        """
        Intialize a new instance of the notebooklet class.

        Parameters
        ----------
        data_providers : DataProviders, Optional
            Optional DataProviders instance to query data.
            Most classes require this.

        """
        super().__init__(data_providers, **kwargs)

        self.asn_selector: Optional[nbwidgets.SelectSubset] = None
        self.flow_index_data = pd.DataFrame()

    # pylint: disable=too-many-branches
    @set_text(docs=_CELL_DOCS, key="run")  # noqa: MC0001
    def run(  # noqa: MC0001, C901
        self,
        value: Any = None,
        data: Optional[pd.DataFrame] = None,
        timespan: Optional[TimeSpan] = None,
        options: Optional[Iterable[str]] = None,
        **kwargs,
    ) -> NetworkFlowResult:
        """
        Return host summary data.

        Parameters
        ----------
        value : str
            Host entity, hostname or host IP Address
        data : Optional[pd.DataFrame], optional
            Not used, by default None
        timespan : TimeSpan
            Timespan over which operations such as queries will be
            performed, by default None.
            This can be a TimeStamp object or another object that
            has valid `start`, `end`, or `period` attributes.
        options : Optional[Iterable[str]], optional
            List of options to use, by default None
            A value of None means use default options.
            Options prefixed with "+" will be added to the default options.
            To see the list of available options type `help(cls)` where
            "cls" is the notebooklet class or an instance of this class.

        Other Parameters
        ----------------
        start : Union[datetime, datelike-string]
            Alternative to specifying timespan parameter.
        end : Union[datetime, datelike-string]
            Alternative to specifying timespan parameter.

        Returns
        -------
        HostNetworkResult
            Result object with attributes for each result type.

        Raises
        ------
        MsticnbMissingParameterError
            If required parameters are missing

        """
        super().run(
            value=value, data=data, timespan=timespan, options=options, **kwargs
        )

        if not value:
            raise MsticnbMissingParameterError("value")
        if not timespan:
            raise MsticnbMissingParameterError("timespan.")

        result = NetworkFlowResult(
            notebooklet=self, description=self.metadata.description, timespan=timespan
        )

        if isinstance(value, entities.Host):
            host_name = value.HostName
            result.host_entity = value
            host_ip = None
        else:
            host_ip, host_name = _ip_or_hostname(value)
            result.host_entity = _create_host_entity(
                host_name=host_name, host_ip=host_ip
            )

        result.description = (
            f"Network flow summary for {host_name or host_ip or 'unknown'}"
        )  # type: ignore

        flow_df = _get_az_net_flows(
            self.query_provider, self.timespan, host_ip, host_name
        )
        result.network_flows = flow_df

        if "resolve_host" in self.options or not hasattr(
            result.host_entity, "IpAddress"
        ):
            result.host_entity = _get_host_details(
                qry_prov=self.query_provider, host_entity=result.host_entity
            )
        if "plot_flows" in self.options:
            result.plot_flows_by_protocol = _plot_flows_by_protocol(flow_df)
            result.plot_flows_by_direction = _plot_flows_by_direction(flow_df)
        if "plot_flow_values" in self.options:
            result.plot_flow_values = _plot_flow_values(flow_df)
        flow_index = None
        if "flow_summary" in self.options:
            flow_index = _extract_flow_ips(flow_df)
            result.flow_index = _get_flow_index(flow_index)
            if not self.silent:
                display(result.flow_index)
            result.flow_summary = _get_flow_summary(flow_index)
            if not self.silent:
                display(result.flow_summary)
            result.flow_index_data = flow_index
        if "geo_map" in self.options and flow_index is not None:
            geo_lookup = self.get_provider("geolitelookup") or self.get_provider(
                "ipstacklookup"
            )
            result.geo_map = _display_geo_map_all(
                flow_index=flow_index,
                ip_locator=geo_lookup,
                host_entity=result.host_entity,
            )
            if not self.silent:
                display(result.geo_map)

        nb_markdown("Select ASNs to examine using select_asns()")
        nb_markdown(
            "Lookup threat intel for IPs from selected ASNs using"
            + " lookup_ti_for_asn_ips()"
        )
        nb_markdown("Display Geolocation of threats with show_selected_asn_map()")
        nb_markdown("For usage type 'help(NetworkFlowSummary.function_name)'")

        # pylint: disable=attribute-defined-outside-init
        # (defined in parent class)
        self._last_result = result
        return self._last_result

    def select_asns(self):
        """Show interactive selector to choose which ASNs to process."""
        if not self._last_result or self._last_result.flow_summary is None:
            print(
                "Please use 'run' with 'flow_summary' option before using",
                "this method.",
            )
            return
        self.asn_selector = _select_asn_subset(
            flow_sum_df=self._last_result.flow_summary,
            host_entity=self._last_result.host_entity,
        )
        display(self.asn_selector)

    def lookup_ti_for_asn_ips(self):
        """Lookup Threat Intel data for IPs of selected ASNs."""
        if (
            not self._last_result
            or self._last_result.flow_summary is None
            or not self.asn_selector
        ):
            print(
                "Please use 'run()' with 'flow_summary' option before using",
                "this method.\n",
                "Then call 'select_asns()' to select the ASNs to lookup.\n",
                "Then call 'lookup_ti_for_asn_ips()'.",
            )
            return

        selected_ips = _get_ips_from_selected_asn(
            flow_sum_df=self._last_result.flow_summary, select_asn=self.asn_selector
        )
        self._last_result.ti_results = _lookup_ip_ti(
            flows_df=self._last_result.flow_index_data,
            selected_ips=selected_ips,
            ti_lookup=self.get_provider("tilookup"),
        )

    def show_selected_asn_map(self) -> foliummap.FoliumMap:
        """
        Display map of IP locations for selected ASNs.

        Returns
        -------
        FoliumMap
            msticpy Folium instance.

        """
        if (
            not self._last_result
            or self._last_result.flow_summary is None
            or not self.asn_selector
        ):
            print(
                "Please use 'run()' with 'flow_summary' option before using",
                "this method. Then run 'select_asns()' (and optionally)",
                "'lookup_ti_for_asn_ips()'.",
                "\nThen call 'show_selected_asn_map()'",
            )
            return None
        geo_lookup = self.get_provider("geolitelookup") or self.get_provider(
            "ipstacklookup"
        )
        geo_map = _display_geo_map(
            flow_index=self._last_result.flow_summary,
            ip_locator=geo_lookup,
            host_entity=self._last_result.host_entity,
            ti_results=self._last_result.ti_results,
            select_asn=self.asn_selector,
        )
        if self.silent:
            display(geo_map)
        return geo_map


# %%
# Helper functions
def _ip_or_hostname(search_value) -> Tuple[Any, Any]:
    search_value = search_value.strip()
    try:
        ip_address(search_value)
        return search_value, None
    except ValueError:
        return None, search_value


def _create_host_entity(host_name=None, host_ip=None):
    host_entity = entities.Host()
    if host_name:
        host_entity.HostName = host_name
    if host_ip:
        ip_entity = entities.IpAddress(Address=host_ip)
        host_entity.IpAddress = ip_entity
    return host_entity


def _get_host_details(qry_prov, host_entity):
    host_ip = getattr(host_entity, "IpAddress", None)
    host_name = getattr(host_entity, "HostName", None)

    host_entity = get_heartbeat(qry_prov=qry_prov, host_ip=host_ip, host_name=host_name)
    get_aznet_topology(
        qry_prov=qry_prov, host_ip=host_ip, host_entity=host_entity, host_name=host_name
    )
    return host_entity


# %%
# Get network flows
def _get_az_net_flows(qry_prov, timespan, ip_addr, hostname):
    nb_data_wait("AzureNetworkAnalytics")
    if ip_addr:
        flow_df = qry_prov.Network.list_azure_network_flows_by_ip(
            timespan, ip_address_list=ip_addr
        )
    else:
        flow_df = qry_prov.Network.list_azure_network_flows_by_host(
            timespan, host_name=hostname
        )

    flow_df["TotalAllowedFlows"] = (
        flow_df["AllowedOutFlows"] + flow_df["AllowedInFlows"]
    )
    return flow_df


# %%
# Plot flows
@set_text(docs=_CELL_DOCS, key="plot_flows_by_protocol")
def _plot_flows_by_protocol(flow_df):
    return display_timeline(
        data=flow_df,
        group_by="L7Protocol",
        title="Network Flows by Protocol",
        time_column="FlowStartTime",
        source_columns=["FlowType", "AllExtIPs", "L7Protocol", "FlowDirection"],
        height=300,
        legend="right",
        yaxis=True,
    )


@set_text(docs=_CELL_DOCS, key="plot_flows_by_direction")
def _plot_flows_by_direction(flow_df):
    return display_timeline(
        data=flow_df,
        group_by="FlowDirection",
        title="Network Flows by Direction",
        time_column="FlowStartTime",
        source_columns=["FlowType", "AllExtIPs", "L7Protocol", "FlowDirection"],
        height=300,
        legend="right",
        yaxis=True,
    )


# %%
# Plot flow values
@set_text(docs=_CELL_DOCS, key="plot_flow_values")
def _plot_flow_values(flow_df, related_alert=None):
    return display_timeline_values(
        data=flow_df,
        group_by="L7Protocol",
        source_columns=[
            "FlowType",
            "AllExtIPs",
            "L7Protocol",
            "FlowDirection",
            "TotalAllowedFlows",
        ],
        time_column="FlowStartTime",
        title="Network flows by Layer 7 Protocol",
        y="TotalAllowedFlows",
        ref_event=related_alert,
        legend="right",
        height=500,
        kind=["vbar", "circle"],
    )


# %%
# Extract source and dest IPs
def _extract_flow_ips(flow_df):
    cols = [
        "VMName",
        "VMIPAddress",
        "PublicIPs",
        "SrcIP",
        "DestIP",
        "L4Protocol",
        "L7Protocol",
        "DestPort",
        "FlowDirection",
        "AllExtIPs",
        "TotalAllowedFlows",
    ]
    flow_index = flow_df[cols].copy()

    def get_source_ip(row):
        if row.FlowDirection == "O":
            return row.VMIPAddress or row.SrcIP
        return row.AllExtIPs or row.DestIP

    def get_dest_ip(row):
        if row.FlowDirection == "O":
            return row.AllExtIPs or row.DestIP
        return row.VMIPAddress or row.SrcIP

    flow_index["source"] = flow_index.apply(get_source_ip, axis=1)
    flow_index["dest"] = flow_index.apply(get_dest_ip, axis=1)

    return flow_index


@set_text(docs=_CELL_DOCS, key="get_flow_index")
def _get_flow_index(flow_summary_df):
    return (
        flow_summary_df[
            ["source", "dest", "L7Protocol", "FlowDirection", "TotalAllowedFlows"]
        ]
        .groupby(["source", "dest", "L7Protocol", "FlowDirection"])
        .sum()
        .reset_index()
        .style.bar(subset=["TotalAllowedFlows"], color="#d65f5f")
    )


# %%
# Flow Summary and Whois lookup
@set_text(docs=_CELL_DOCS, key="get_flow_summary")
def _get_flow_summary(flow_index):
    flows_df = (
        flow_index[
            ["source", "dest", "L7Protocol", "FlowDirection", "TotalAllowedFlows"]
        ]
        .groupby(["source", "dest", "L7Protocol", "FlowDirection"])
        .sum()
        .reset_index()
    )

    num_ips = len(flows_df["source"].unique()) + len(flows_df["dest"].unique())
    nb_markdown(f"Found {num_ips} unique IP Addresses.")

    nb_data_wait("Whois")
    asn_columns = []
    if flows_df["dest"].apply(get_ip_type).unique() != ["Private"]:
        flows_df = get_whois_df(
            flows_df,
            ip_column="dest",
            asn_col="DestASN",
            whois_col="DestASNFull",
            show_progress=True,
        )
        asn_columns.append("DestASN")
    if flows_df["source"].apply(get_ip_type).unique() != ["Private"]:
        flows_df = get_whois_df(
            flows_df,
            ip_column="source",
            asn_col="SourceASN",
            whois_col="SourceASNFull",
            show_progress=True,
        )
        asn_columns.append("SourceASN")

    if isinstance(flows_df, pd.DataFrame) and not flows_df.empty:
        return (
            flows_df.groupby(asn_columns)
            .agg(
                TotalAllowedFlows=pd.NamedAgg(
                    column="TotalAllowedFlows", aggfunc="sum"
                ),
                L7Protocols=pd.NamedAgg(
                    column="L7Protocol", aggfunc=lambda x: x.unique().tolist()
                ),
                source_ips=pd.NamedAgg(
                    column="source", aggfunc=lambda x: x.unique().tolist()
                ),
                dest_ips=pd.NamedAgg(
                    column="dest", aggfunc=lambda x: x.unique().tolist()
                ),
            )
            .reset_index()
        )

    return None


# %%
# ASN Selection
def _get_source_host_asns(host_entity):
    host_ips = getattr(host_entity, "PublicIpAddresses", [])
    host_ips.append(getattr(host_entity, "IpAddress", None))
    host_asns = []
    for ip_entity in host_ips:
        if get_ip_type(ip_entity.Address) == "Public":
            whois_result = get_whois_info(ip_entity)
            if hasattr(whois_result, "properties"):
                asn_name = whois_result.name
            else:
                asn_name = whois_result[0]
            host_asns.append(asn_name)
    return host_asns


@set_text(docs=_CELL_DOCS, key="select_asn_subset")
def _select_asn_subset(flow_sum_df, host_entity):
    our_host_asns = _get_source_host_asns(host_entity)
    all_asns: List[str] = []
    other_asns: List[str] = []

    # Select the ASNs in the 25th percentile (lowest number of flows)
    quant_25pc = flow_sum_df["TotalAllowedFlows"].quantile(q=[0.25]).iat[0]
    quant_25pc_df = flow_sum_df[flow_sum_df["TotalAllowedFlows"] <= quant_25pc]

    if "DestASN" in flow_sum_df.columns:
        all_asns.extend(flow_sum_df["DestASN"].unique())
        other_asns.extend(quant_25pc_df["DestASN"].unique())
    if "SourceASN" in flow_sum_df.columns:
        all_asns.extend(flow_sum_df["SourceASN"].unique())
        other_asns.extend(quant_25pc_df["SourceASN"].unique())
    all_asns = set(all_asns) - {"private address"}
    other_asns = set(other_asns) - set(our_host_asns)
    return nbwidgets.SelectSubset(source_items=all_asns, default_selected=other_asns)


# %%
# Lookup ASN IPs with TILookup
def _get_ips_from_selected_asn(flow_sum_df, select_asn):
    dest_ips: Set[str] = set()
    src_ips: Set[str] = set()
    if "DestASN" in flow_sum_df.columns:
        dest_ips = set(
            chain.from_iterable(
                flow_sum_df[flow_sum_df["DestASN"].isin(select_asn.selected_items)][
                    "dest_ips"
                ]
            )
        )
    if "SourceASN" in flow_sum_df.columns:
        src_ips = set(
            chain.from_iterable(
                flow_sum_df[flow_sum_df["SourceASN"].isin(select_asn.selected_items)][
                    "source_ips"
                ]
            )
        )
    selected_ips = dest_ips | src_ips
    nb_markdown(f"{len(selected_ips)} unique IPs in selected ASNs")
    return selected_ips


@set_text(docs=_CELL_DOCS, key="lookup_ip_ti")
def _lookup_ip_ti(flows_df, ti_lookup, selected_ips):
    def ti_check_ser_sev(severity, threshold):
        threshold = ResultSeverity.parse(threshold)
        return severity.apply(lambda x: ResultSeverity.parse(x) >= threshold)

    # Add the IoCType to save cost of inferring each item
    nb_data_wait("Threat Intelligence")
    selected_ip_dict = {ip: "ipv4" for ip in selected_ips}
    ti_results = ti_lookup.lookup_iocs(data=selected_ip_dict)

    nb_markdown(f"{len(ti_results)} TI results received.")
    if ti_results.empty:
        return pd.DataFrame(columns=["Ioc"])

    ti_results_pos = ti_results[ti_check_ser_sev(ti_results["Severity"], 1)]
    nb_markdown(f"{len(ti_results_pos)} positive results found.")

    if not ti_results_pos.empty:
        src_pos = flows_df.merge(ti_results_pos, left_on="source", right_on="Ioc")
        dest_pos = flows_df.merge(ti_results_pos, left_on="dest", right_on="Ioc")
        ti_ip_results = pd.concat([src_pos, dest_pos])
        nb_warn("Positive Threat Intel Results found for the following flows")
        nb_markdown(
            "Please examine these IP flows using the IP Explorer notebook.",
            "bold, large",
        )
        return ti_ip_results
    return pd.DataFrame(columns=["Ioc"])


# %%
# display GeoLocations of IPs
def _format_ip_entity(ip_loc, row, ip_col):
    ip_entity = entities.IpAddress(Address=row[ip_col])
    ip_loc.lookup_ip(ip_entity=ip_entity)
    if "L7Protocol" in row:
        ip_entity.AdditionalData["protocol"] = row.L7Protocol
    if "severity" in row:
        ip_entity.AdditionalData["threat severity"] = row["severity"]
    if "Details" in row:
        ip_entity.AdditionalData["threat details"] = row["Details"]
    return ip_entity


# pylint: disable=too-many-branches
@set_text(docs=_CELL_DOCS, key="display_geo_map_all")
def _display_geo_map_all(flow_index, ip_locator, host_entity):
    folium_map = foliummap.FoliumMap(zoom_start=4)
    if flow_index is None or flow_index.empty:
        nb_markdown("No network flow data available.")
        return None

    # Get the flow records for all flows not in the TI results
    selected_out = flow_index

    if selected_out.empty:
        ips_out = []
    else:
        nb_data_wait("IP Geolocation")
        ips_out = list(
            selected_out.apply(
                lambda x: _format_ip_entity(ip_locator, x, "dest"), axis=1
            )
        )

    selected_in = flow_index
    if selected_in.empty:
        ips_in = []
    else:
        nb_data_wait("IP Geolocation")
        ips_in = list(
            selected_in.apply(
                lambda x: _format_ip_entity(ip_locator, x, "source"), axis=1
            )
        )

    icon_props = {"color": "green"}
    host_ips = getattr(host_entity, "PublicIpAddresses", [])
    host_ip = getattr(host_entity, "IpAddress", None)
    if host_ip:
        host_ips.append(host_ip)
    if host_ips:
        for ips in host_ips:
            ips.AdditionalData["host"] = host_entity.HostName or "unknown hostname"
        folium_map.add_ip_cluster(ip_entities=host_ips, **icon_props)
    icon_props = {"color": "blue"}
    folium_map.add_ip_cluster(ip_entities=ips_out, **icon_props)
    icon_props = {"color": "purple"}
    folium_map.add_ip_cluster(ip_entities=ips_in, **icon_props)
    folium_map.center_map()
    return folium_map


# pylint: enable=too-many-branches


# pylint: disable=too-many-branches, too-many-locals
@set_text(docs=_CELL_DOCS, key="display_geo_map")
def _display_geo_map(flow_index, ip_locator, host_entity, ti_results, select_asn):
    folium_map = foliummap.FoliumMap(zoom_start=4)
    if flow_index is None or flow_index.empty:
        nb_markdown("No network flow data available.")
        return None

    ips_in: List[str] = []
    ips_out: List[str] = []
    # Get the flow records for all flows not in the TI results
    if "DestASN" in flow_index.columns:
        selected_out = flow_index[flow_index["DestASN"].isin(select_asn.selected_items)]
        sel_out_exp = selected_out.explode("dest_ips")
        sel_out_exp = sel_out_exp[~sel_out_exp["dest_ips"].isin(ti_results["Ioc"])]

        if not sel_out_exp.empty:
            nb_data_wait("IP Geolocation")
            ips_out = list(
                sel_out_exp.apply(
                    lambda x: _format_ip_entity(ip_locator, x, "dest_ips"), axis=1
                )
            )

    if "SourceASN" in flow_index.columns:
        selected_in = flow_index[
            flow_index["SourceASN"].isin(select_asn.selected_items)
        ]
        sel_in_exp = selected_in.explode("source_ips")
        sel_in_exp = sel_in_exp[~sel_in_exp["source_ips"].isin(ti_results["Ioc"])]

        if not sel_in_exp.empty:
            nb_data_wait("IP Geolocation")
            ips_in = list(
                sel_in_exp.apply(
                    lambda x: _format_ip_entity(ip_locator, x, "source_ips"), axis=1
                )
            )

    icon_props = {"color": "green"}
    host_ips = getattr(host_entity, "PublicIpAddresses", [])
    host_ip = getattr(host_entity, "IpAddress", None)
    if host_ip:
        host_ips.append(host_ip)
    if host_ips:
        for ip_addr in host_ips:
            ip_addr.AdditionalData["host"] = host_entity.HostName or "unknown hostname"
        folium_map.add_ip_cluster(ip_entities=host_ips, **icon_props)
    icon_props = {"color": "blue"}
    folium_map.add_ip_cluster(ip_entities=ips_out, **icon_props)
    icon_props = {"color": "purple"}
    folium_map.add_ip_cluster(ip_entities=ips_in, **icon_props)
    if not (ti_results is None or ti_results.empty):
        ips_threats = list(
            ti_results.apply(lambda x: _format_ip_entity(ip_locator, x, "Ioc"), axis=1)
        )
        icon_props = {"color": "red"}
        folium_map.add_ip_cluster(ip_entities=ips_threats, **icon_props)
    folium_map.center_map()

    return folium_map



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""IP Address Summary notebooklet."""
import json
from contextlib import suppress
from ipaddress import IPv4Address, IPv6Address, ip_address
from json import JSONDecodeError
from typing import Any, Dict, Iterable, List, Optional, Union

import numpy as np
import pandas as pd
from bokeh.models import LayoutDOM
from msticpy.common.exceptions import MsticpyException
from msticpy.common.timespan import TimeSpan
from msticpy.datamodel.entities import GeoLocation, Host, IpAddress

try:
    from msticpy import nbwidgets
    from msticpy.context.ip_utils import get_ip_type
    from msticpy.context.ip_utils import ip_whois as get_whois_info
    from msticpy.vis.timeline import display_timeline
except ImportError:
    # Fall back to msticpy locations prior to v2.0.0
    from msticpy.nbtools import nbwidgets
    from msticpy.nbtools.timeline import display_timeline
    from msticpy.sectools.ip_utils import get_ip_type, get_whois_info

    # pylint: disable=unused-import
    from msticpy.vis import mp_pandas_plot  # noqa: F401

from .... import nb_metadata
from ...._version import VERSION
from ....common import (
    MsticnbMissingParameterError,
    df_has_data,
    nb_data_wait,
    nb_display,
    nb_markdown,
    set_text,
)
from ....nblib.azsent.alert import browse_alerts
from ....nblib.azsent.host import populate_host_entity
from ....nblib.iptools import convert_to_ip_entities
from ....notebooklet import NBMetadata, Notebooklet, NotebookletResult

__version__ = VERSION
__author__ = "Ian Hellen"


# Read module metadata from YAML
_CLS_METADATA: NBMetadata
_CELL_DOCS: Dict[str, Any]
_CLS_METADATA, _CELL_DOCS = nb_metadata.read_mod_metadata(__file__, __name__)


# pylint: disable=too-few-public-methods, too-many-instance-attributes, too-many-lines
# Rename this class
class IpSummaryResult(NotebookletResult):
    """
    IPSummary Results.

    Attributes
    ----------
    ip_str : str
        The input IP address as a string.
    ip_address : Optional[Union[IPv4Address, IPv6Address]]
        Ip Address Python object
    ip_entity : IpAddress
        IpAddress entity
    ip_origin : str
        "External" or "Internal"
    host_entities : Host
        Host entity or entities associated with IP Address
    ip_type : str
        IP address type - "Public", "Private", etc.
    geoip : Optional[Dict[str, Any]]
        Geo location information as a dictionary.
    location : Optional[GeoLocation]
        Location entity context object.
    whois : pd.DataFrame
        WhoIs information for IP Address
    whois_nets : pd.DataFrame
        List of networks definitions from WhoIs data
    heartbeat : pd.DataFrame
        Heartbeat record for IP Address or host
    az_network_if : pd.DataFrame
        Azure NSG analytics interface record, if available
    vmcomputer : pd.DataFrame
        VMComputer latest record
    az_network_flows : pd.DataFrame
        Azure NSG flows for IP, if available
    az_network_flows_timeline: LayoutDOM
        Azure NSG flows timeline, if data is available
    aad_signins : pd.DataFrame = None
        AAD signin activity
    azure_activity : pd.DataFrame = None
        Azure Activity log entries
    azure_activity_summary : pd.DataFrame = None
        Azure Activity (AAD and Az Activity) summarized view
    office_activity : pd.DataFrame = None
        Office 365 activity
    common_security : pd.DataFrame
        Common Security Log entries for source IP
    related_bookmarks : pd.DataFrame
        Bookmarks related to IP Address
    alert_timeline : LayoutDOM
        Timeline plot of alerts
    ti_results: pd.DataFrame
        Threat intel lookup results
    passive_dns: pd.DataFrame
        Passive DNS lookup results
    self.host_logons : pd.DataFrame
        Hosts with logons from this IP Address
    self.related_accounts : pd.DataFrame
        Accounts with activity related to this IP Address
    self.associated_hosts : pd.DataFrame
        Hosts using this IP Address
    self.device_info : pd.DataFrame
        Device info of hosts using this IP Address
    self.network_connections : pd.DataFrame = None
        Network connections to/from this IP on other devices

    """

    def __init__(
        self,
        description: Optional[str] = None,
        timespan: Optional[TimeSpan] = None,
        notebooklet: Optional["Notebooklet"] = None,
    ):
        """
        Create new IPSummaryResult result instance.

        Parameters
        ----------
        description : Optional[str], optional
            Result description, by default None
        timespan : Optional[TimeSpan], optional
            TimeSpan for the results, by default None
        notebooklet : Optional[, optional
            Originating notebooklet, by default None

        """
        super().__init__(description, timespan, notebooklet)
        self.description: str = "IP Address Summary"

        # Add attributes as needed here.
        # Make sure they are documented in the Attributes section
        # above.
        self.ip_str: str = ""
        self.ip_address: Optional[Union[IPv4Address, IPv6Address]] = None
        self.ip_entity: IpAddress = None
        self.ip_origin: str = "External"
        self.ip_type: str = "Public"
        self.host_entities: List[Host] = []
        self.geoip: Optional[Dict[str, Any]] = None
        self.location: Optional[GeoLocation] = None
        self.whois: Optional[pd.DataFrame] = None
        self.whois_nets: Optional[pd.DataFrame] = None
        self.heartbeat: Optional[pd.DataFrame] = None
        self.az_network_if: Optional[pd.DataFrame] = None
        self.vmcomputer: Optional[pd.DataFrame] = None
        self.az_network_flows: Optional[pd.DataFrame] = None
        self.az_network_flow_summary: Optional[pd.DataFrame] = None
        self.az_network_flows_timeline: Optional[LayoutDOM] = None
        self.aad_signins: Optional[pd.DataFrame] = None
        self.azure_activity: Optional[pd.DataFrame] = None
        self.azure_activity_summary: Optional[pd.DataFrame] = None
        self.office_activity: Optional[pd.DataFrame] = None
        self.common_security: Optional[pd.DataFrame] = None
        self.related_alerts: Optional[pd.DataFrame] = None
        self.related_bookmarks: Optional[pd.DataFrame] = None
        self.alert_timeline: Optional[LayoutDOM] = None
        self.ti_results: Optional[pd.DataFrame] = None
        self.passive_dns: Optional[pd.DataFrame] = None
        self.host_logons: Optional[pd.DataFrame] = None
        self.related_accounts: Optional[pd.DataFrame] = None
        self.associated_hosts: Optional[pd.DataFrame] = None
        self.device_info: Optional[pd.DataFrame] = None
        self.network_connections: Optional[pd.DataFrame] = None


# pylint: enable=too-few-public-methods, too-many-instance-attributes


class IpAddressSummary(Notebooklet):
    """
    IP Address Summary Notebooklet class.

    Queries and displays summary information about an IP address, including:

    - Basic IP address properties
    - IpAddress entity (and Host entity, if a host could be associated)
    - WhoIs and Geo-location
    - Azure activity and network data (optional)
    - Office activity summary (optional)
    - Threat intelligence reports
    - Related alerts and hunting bookmarks

    """

    # assign metadata from YAML to class variable
    metadata = _CLS_METADATA
    __doc__ = nb_metadata.update_class_doc(__doc__, metadata)
    _cell_docs = _CELL_DOCS

    # pylint: disable=too-many-branches, too-many-statements
    @set_text(docs=_CELL_DOCS, key="run")  # noqa: MC0001
    def run(  # noqa: MC0001,C901
        self,
        value: Any = None,
        data: Optional[pd.DataFrame] = None,
        timespan: Optional[TimeSpan] = None,
        options: Optional[Iterable[str]] = None,
        **kwargs,
    ) -> IpSummaryResult:
        """
        Return IP Address activity summary.

        Parameters
        ----------
        value : str
            IP Address - The key for searches
        data : Optional[pd.DataFrame], optional
            Not supported for this notebooklet.
        timespan : TimeSpan
            Timespan for queries
        options : Optional[Iterable[str]], optional
            List of options to use, by default None.
            A value of None means use default options.
            Options prefixed with "+" will be added to the default options.
            To see the list of available options type `help(cls)` where
            "cls" is the notebooklet class or an instance of this class.

        Returns
        -------
        IpSummaryResult
            Result object with attributes for each result type.

        Raises
        ------
        MsticnbMissingParameterError
            If required parameters are missing

        """
        # This line use logic in the superclass to populate options
        # (including default options) into this class.
        super().run(
            value=value, data=data, timespan=timespan, options=options, **kwargs
        )

        if not value:
            raise MsticnbMissingParameterError("value")
        if not timespan:
            raise MsticnbMissingParameterError("timespan.")

        # Create a result class
        result = IpSummaryResult(
            notebooklet=self, description=self.metadata.description, timespan=timespan
        )

        result.ip_str = value
        result.ip_address = ip_address(value)
        result.ip_type = get_ip_type(value)
        result.ip_entity = IpAddress(Address=value)
        nb_markdown(f"{value}, ip address type: {result.ip_type}")

        geo_lookup = self.get_provider("geolitelookup") or self.get_provider(
            "ipstacklookup"
        )
        # Lookup basic host info
        if "device_info" in self.options and self.check_table_exists(
            "DeviceNetworkInfo"
        ):
            _get_device_info(
                qry_prov=self.query_provider,
                src_ip=value,
                result=result,
                geo_lookup=geo_lookup,
            )
        if not df_has_data(result.device_info):
            self.options.extend(["az_net_if", "heartbeat", "vmcomputer"])
            self.options = list(set(self.options))
        if "az_net_if" in self.options and self.check_table_exists(
            "AzureNetworkAnalytics_CL"
        ):
            _get_az_net_if(qry_prov=self.query_provider, src_ip=value, result=result)
        if "heartbeat" in self.options and self.check_table_exists("Heartbeat"):
            _get_heartbeat(qry_prov=self.query_provider, src_ip=value, result=result)
        if "vmcomputer" in self.options and self.check_table_exists("VMComputer"):
            _get_vmcomputer(qry_prov=self.query_provider, src_ip=value, result=result)
        _populate_host_entity(result, geo_lookup=geo_lookup)
        # Future: (ianhelle) - _normalize_host_entities(result)

        # Alerts and bookmarks
        if "alerts" in self.options:
            self._get_related_alerts(src_ip=value, result=result, timespan=timespan)
        if "bookmarks" in self.options:
            result = self._get_related_bookmarks(
                src_ip=value, result=result, timespan=timespan
            )
        # Azure NSG netflow
        if "az_netflow" in self.options:
            self._get_azure_netflow(src_ip=value, result=result, timespan=timespan)
            if df_has_data(result.az_network_flows):
                result.az_network_flow_summary = _summarize_netflows(
                    result.az_network_flows
                )
                nb_display()
        if "az_activity" in self.options or "related_accounts" in self.options:
            self._get_azure_activity(src_ip=value, result=result, timespan=timespan)
            _summarize_azure_activity(result)
        if "office_365" in self.options:
            self._get_office_activity(src_ip=value, result=result, timespan=timespan)
        if "common_security" in self.options:
            _get_common_security_log(
                qry_prov=self.query_provider,
                src_ip=value,
                result=result,
                timespan=timespan,
            )

        if "host_logons" in self.options:
            _get_host_logons(
                qry_prov=self.query_provider,
                src_ip=value,
                result=result,
                timespan=timespan,
            )
        if "related_accounts" in self.options:
            _get_related_accounts(result=result)

        if "device_network" in self.options and self.check_table_exists(
            "DeviceNetworkEvents"
        ):
            _get_device_net_connections(
                qry_prov=self.query_provider, src_ip=value, result=result
            )

        result.ip_origin = _determine_ip_origin(result)

        if result.ip_type == "Public":
            self._get_public_ip_data(src_ip=value, result=result)

        # Assign the result to the _last_result attribute
        # so that you can get to it without having to re-run the operation
        self._last_result = result  # pylint: disable=attribute-defined-outside-init

        nb_markdown("<h3>View the returned results object for more details.</h3>")
        nb_markdown(
            f"Additional methods for this class:<br>{'<br>'.join(self.list_methods())}"
        )
        return self._last_result

    # pylint: disable=too-many-branches, too-many-statements

    def browse_alerts(self) -> nbwidgets.SelectAlert:
        """Return alert browser/viewer."""
        if self.check_valid_result_data("related_alerts"):
            return browse_alerts(self._last_result, "related_alerts")
        return None

    def display_alert_timeline(self):
        """Display the alert timeline."""
        if self.check_valid_result_data("related_alerts"):
            return display_timeline(
                data=self._last_result.related_alerts,
                title="Alerts",
                source_columns=["AlertName"],
                height=300,
                hide=True,
            )
        return None

    def browse_ti_results(self):
        """Display Threat intel results."""
        if self.check_valid_result_data("ti_results"):
            ti_lookup = self.get_provider("tilookup")
            return ti_lookup.browse_results(self._last_result.ti_results)
        return None

    def netflow_by_protocol(
        self,
    ) -> Optional[LayoutDOM]:
        """Display netflows grouped by protocol."""
        if not self.check_valid_result_data("az_network_flows"):
            return None
        return _plot_netflow_by_protocol(self._last_result)

    def netflow_total_by_protocol(
        self,
    ) -> Optional[LayoutDOM]:
        """Display netflows grouped by protocol."""
        if not self.check_valid_result_data("az_network_flows"):
            return None
        return _plot_netflow_values_by_protocol(self._last_result)

    def netflow_by_direction(
        self,
    ) -> Optional[LayoutDOM]:
        """Display netflows grouped by direction."""
        if not self.check_valid_result_data("az_network_flows"):
            return None
        return _plot_netflow_by_direction(self._last_result)

    @set_text(docs=_CELL_DOCS, key="get_public_ip_data")
    def _get_public_ip_data(self, src_ip, result):
        """Retrieve data for public IP."""
        _get_whois(src_ip, result)
        nb_markdown("WhoIs data")
        nb_display(pd.DataFrame(result.whois.iloc[0]).T)

        # GeoIP
        if "geoip" in self.options:
            geo_lookup = self.get_provider("geolitelookup") or self.get_provider(
                "ipstacklookup"
            )
            if geo_lookup:
                _get_geoip_data(geo_lookup, src_ip, result)

        ti_provider = self.get_provider(provider_name="tilookup")
        # TI Lookup
        if result.ip_origin == "External" or "ti" in self.options:
            _get_ti_data(ti_provider, src_ip, result)

        # Passive DNS
        if (
            result.ip_origin == "External" or "passive_dns" in self.options
        ) and isinstance(result.ip_address, IPv4Address):
            _get_passv_dns(ti_provider, src_ip, result)

    @set_text(docs=_CELL_DOCS, key="get_az_netflow")
    def _get_azure_netflow(self, src_ip, result, timespan):
        """Retrieve Azure netflow and activity events."""
        if self.check_table_exists("AzureNetworkAnalytics_CL"):
            _get_az_netflows(self.query_provider, src_ip, result, timespan)
            _display_df_summary(result.az_network_flows, "Azure NSG network flows")

    @set_text(docs=_CELL_DOCS, key="get_az_activity")
    def _get_azure_activity(self, src_ip, result, timespan):
        """Retrieve Azure netflow and activity events."""
        if self.check_table_exists("SigninLogs"):
            nb_data_wait("SigninLogs")
            result.aad_signins = self.query_provider.Azure.list_aad_signins_for_ip(
                timespan, ip_address_list=src_ip
            )
            _display_df_summary(result.aad_signins, "AAD signins")

        if self.check_table_exists("AzureActivity"):
            nb_data_wait("AzureActivity")
            result.azure_activity = (
                self.query_provider.Azure.list_azure_activity_for_ip(
                    timespan, ip_address_list=src_ip
                )
            )
            _display_df_summary(result.azure_activity, "Azure Activity")

    @set_text(docs=_CELL_DOCS, key="get_office_activity")
    def _get_office_activity(self, src_ip, result, timespan):
        """Retrieve Office activity data."""
        if self.check_table_exists("OfficeActivity"):
            nb_data_wait("OfficeActivity")
            office_activity = self.query_provider.Office365.list_activity_for_ip(
                timespan, ip_address_list=src_ip
            )
            if df_has_data(office_activity):
                result.office_activity = (
                    office_activity.rename(columns={"UserId": "Account"})
                    .groupby(["Account", "OfficeWorkload", "Operation"])
                    .agg(
                        OperationCount=pd.NamedAgg(column="Type", aggfunc="count"),
                        OperationTypes=pd.NamedAgg(
                            column="Operation", aggfunc=lambda x: x.unique().tolist()
                        ),
                        FirstOperation=pd.NamedAgg(
                            column="TimeGenerated", aggfunc="min"
                        ),
                        LastOperation=pd.NamedAgg(
                            column="TimeGenerated", aggfunc="max"
                        ),
                    )
                )
            _display_df_summary(result.office_activity, "Office 365 operations")
            if df_has_data(result.office_activity):
                nb_display(result.office_activity)

    @set_text(docs=_CELL_DOCS, key="get_related_alerts")
    def _get_related_alerts(self, src_ip, result, timespan):
        """Get any related alerts for `src_ip`."""
        nb_data_wait("RelatedAlerts")
        result.related_alerts = self.query_provider.SecurityAlert.list_alerts_for_ip(
            timespan, source_ip_list=src_ip
        )
        _display_df_summary(result.related_alerts, "related alerts")
        if df_has_data(result.related_alerts):
            nb_markdown(
                "Use `browse_alerts` and `display_alert_timeline` to view these."
            )

    @set_text(docs=_CELL_DOCS, key="get_related_alerts")
    def _get_related_bookmarks(
        self, src_ip, result, timespan: TimeSpan
    ) -> pd.DataFrame:
        nb_data_wait("Bookmarks")
        result.related_bookmarks = (
            self.query_provider.AzureSentinel.list_bookmarks_for_entity(  # type: ignore
                timespan, entity_id=src_ip
            )
        )
        _display_df_summary(result.related_bookmarks, "related bookmarks")
        return result


# %%
# Helper functions
def _display_df_summary(data, description):
    if df_has_data(data):
        nb_markdown(f"{len(data)} {description}.")
    else:
        nb_markdown(f"No events from {description} found.")


def _determine_ip_origin(result):
    return (
        "Internal"
        if (
            result.ip_type == "Private"
            or df_has_data(result.heartbeat)
            or df_has_data(result.az_network_if)
            or df_has_data(result.vmcomputer)
            or df_has_data(result.associated_hosts)
        )
        else "External"
    )


# %%
# Get Azure network flows
def _get_az_netflows(qry_prov, src_ip, result, timespan):
    nb_data_wait("Azure NSG flows")
    result.az_network_flows = qry_prov.Network.list_azure_network_flows_by_ip(
        timespan, ip_address_list=src_ip
    )
    if df_has_data(result.az_network_flows):
        result.az_network_flows["TotalAllowedFlows"] = (
            result.az_network_flows["AllowedOutFlows"]
            + result.az_network_flows["AllowedInFlows"]
        )
        result.az_network_flows_timeline = _plot_netflow_by_protocol(result)


def _plot_netflow_by_protocol(result):
    return result.az_network_flows.mp_plot.timeline(
        group_by="L7Protocol",
        title="Network Flows by Protocol",
        time_column="FlowStartTime",
        source_columns=["FlowType", "AllExtIPs", "L7Protocol", "FlowDirection"],
        height=300,
        legend="right",
        yaxis=True,
    )


def _plot_netflow_values_by_protocol(result):
    return result.az_network_flows.mp_plot.timeline_values(
        group_by="L7Protocol",
        source_columns=[
            "FlowType",
            "AllExtIPs",
            "L7Protocol",
            "FlowDirection",
            "TotalAllowedFlows",
        ],
        time_column="FlowStartTime",
        title="Network flows by Layer 7 Protocol",
        y="TotalAllowedFlows",
        legend="right",
        height=500,
        kind=["vbar", "circle"],
    )


def _plot_netflow_by_direction(result):
    return result.az_network_flows.mp_plot.timeline(
        group_by="FlowDirection",
        title="Network Flows by Direction",
        time_column="FlowStartTime",
        source_columns=["FlowType", "AllExtIPs", "L7Protocol", "FlowDirection"],
        height=300,
        legend="right",
        yaxis=True,
    )


@set_text(docs=_CELL_DOCS, key="netflow_summary")
def _summarize_netflows(data):
    # pylint: disable=unnecessary-lambda
    return (
        data[
            [
                "AllExtIPs",
                "L7Protocol",
                "FlowDirection",
                "TotalAllowedFlows",
                "FlowStartTime",
            ]
        ]
        .groupby(["L7Protocol", "FlowDirection"])
        .agg(
            ExtIPs=pd.NamedAgg(column="AllExtIPs", aggfunc=lambda x: ", ".join(x)),
            ExtIPCount=pd.NamedAgg(column="AllExtIPs", aggfunc="count"),
            FirstFlow=pd.NamedAgg(column="FlowStartTime", aggfunc="min"),
            LastFlow=pd.NamedAgg(column="FlowStartTime", aggfunc="max"),
        )
    )
    # pylint: enable=unnecessary-lambda


# %%
# Azure activity
def _summarize_azure_activity(result):
    az_dfs = []
    if df_has_data(result.aad_signins):
        az_dfs.append(
            result.aad_signins.assign(
                UserPrincipalName=lambda x: x.UserPrincipalName.str.lower(),
                Status=result.aad_signins["ResultDescription"],
            )
            .replace({"Status": ""}, "Success")  # fill in blank (ResultType==0) status
            .rename(
                columns={
                    "OperationName": "Operation",
                    "AppDisplayName": "AppResourceProvider",
                }
            )
        )

    if df_has_data(result.azure_activity):
        az_dfs.append(
            result.azure_activity.assign(
                UserPrincipalName=lambda x: x.Caller.str.lower()
            ).rename(
                columns={
                    "CallerIpAddress": "IPAddress",
                    "OperationName": "Operation",
                    "ResourceProvider": "AppResourceProvider",
                    "Category": "UserType",
                    "ActivityStatusValue": "Status",
                }
            )
        )

    if not az_dfs:
        return

    az_all_data = pd.concat(az_dfs)
    result.azure_activity_summary = az_all_data.groupby(
        [
            "UserPrincipalName",
            "Type",
            "IPAddress",
            "AppResourceProvider",
            "UserType",
            "Status",
        ]
    ).agg(
        OperationCount=pd.NamedAgg(column="Type", aggfunc="count"),
        OperationTypes=pd.NamedAgg(
            column="Operation", aggfunc=lambda x: x.unique().tolist()
        ),
        Resources=pd.NamedAgg(column="ResourceId", aggfunc="nunique"),
        FirstOperation=pd.NamedAgg(column="TimeGenerated", aggfunc="min"),
        LastOperation=pd.NamedAgg(column="TimeGenerated", aggfunc="max"),
    )
    nb_markdown("Summary of Azure activity", "large")
    nb_display(result.azure_activity_summary)


# %%
# Azure heartbeat, interface and VMComputer data
@set_text(docs=_CELL_DOCS, key="get_az_net_if")
def _get_az_net_if(qry_prov, src_ip, result):
    """Get the AzureNetwork topology record for `src_ip`."""
    nb_data_wait("Azure NSG topology")
    # Try to find the interface topology log entry
    result.az_network_if = qry_prov.Network.get_host_for_ip(  # type:ignore
        ip_address=src_ip
    )
    if not df_has_data(result.az_network_if):
        nb_markdown("Could not get Azure NSG network interface record.")


@set_text(docs=_CELL_DOCS, key="get_heartbeat")
def _get_heartbeat(qry_prov, src_ip, result):
    """Get the Heartbeat record for `src_ip`."""
    nb_data_wait("Heartbeat")
    if result.ip_type == "Public":
        result.heartbeat = qry_prov.Network.get_heartbeat_for_ip(ip_address=src_ip)
    elif (
        result.host_entity
        and result.host_entity.HostName
        and result.host_entity.HostName != "unknown"
    ):
        result.heartbeat = qry_prov.Network.get_heartbeat_for_host(
            host_name=result.host_entity.HostName
        )
    if not df_has_data(result.heartbeat):
        nb_markdown("Could not get Azure Heartbeat record.")


@set_text(docs=_CELL_DOCS, key="get_common_security_log")
def _get_common_security_log(qry_prov, src_ip, result, timespan):
    """Get the CommonSecurityLog records for `src_ip`."""
    nb_data_wait("Common Security Log")
    query = """
    CommonSecurityLog
    | where TimeGenerated between (datetime({start}) .. datetime({end}))
    | where SourceIP == '{ip_address}' or DestinationIP == '{ip_address}'
    """
    result_df = qry_prov.exec_query(
        query.format(
            start=timespan.start,
            end=timespan.end,
            ip_address=src_ip,
        )
    )

    if df_has_data(result_df):
        # Log has lots of sparse columns - trim these down as much as
        # possible.
        result.common_security = result_df.replace("", np.nan).dropna(
            how="all", axis="columns"
        )
    else:
        nb_markdown("No CommonSecurityLog records found.")


@set_text(docs=_CELL_DOCS, key="get_vmcomputer")
def _get_vmcomputer(qry_prov, src_ip, result):
    """Get the VMComputer record for `src_ip`."""
    nb_data_wait("VMComputer")
    result.vmcomputer = qry_prov.Azure.get_vmcomputer_for_ip(  # type:ignore
        ip_address=src_ip
    )
    if not df_has_data(result.vmcomputer):
        nb_markdown("Could not get VMComputer record")


_MDE_DEVICE_NET_INFO = """
DeviceNetworkInfo
| where IPAddresses has "{src_ip}"
| mv-expand IPAddresses
| extend IPAddress = tostring(IPAddresses.IPAddress)
| summarize arg_max(TimeGenerated, *) by IPAddress, DeviceName, DeviceId
"""

_MDE_DEVICE_INFO = """
DeviceInfo
| where DeviceId == "{device_id}"
| summarize arg_max(TimeGenerated, *)
"""


@set_text(docs=_CELL_DOCS, key="get_device_info")
def _get_device_info(qry_prov, src_ip, result, geo_lookup):
    nb_data_wait("DeviceInfo")
    result.associated_hosts = qry_prov.exec_query(
        _MDE_DEVICE_NET_INFO.format(src_ip=src_ip)
    )
    if df_has_data(result.associated_hosts):
        device_ids = result.associated_hosts.DeviceId.unique()
        result.device_info = pd.concat(
            [
                qry_prov.exec_query(_MDE_DEVICE_INFO.format(device_id=device_id))
                for device_id in device_ids
            ],
            ignore_index=True,
        )
        result.host_entities = _host_entity_from_dev_info(result, geo_lookup=geo_lookup)
        _display_df_summary(result.associated_hosts, "Defender hosts with matching IP")
        nb_display(result.associated_hosts[["DeviceName"]].drop_duplicates())
        _display_df_summary(result.device_info, "Defender hosts device info")
    nb_markdown("No events from Defender Device info")


def _host_entity_from_dev_info(result, geo_lookup):
    device_info = result.device_info
    host_dict: Dict[str, Host] = {}
    for row in device_info.itertuples():
        host_entity = host_dict.get(row.DeviceName)
        # if we don't already have a host of that name, create one
        if not host_entity:
            if "." in row.DeviceName:
                host, domain = row.DeviceName.split(".", maxsplit=1)
            else:
                host = row.DeviceName
                domain = None
            host_entity = Host(
                HostName=host,
                DnsDomain=domain,
                IsDomainJoined=row.IsAzureADJoined,
            )
            host_entity.DeviceIds = []
            host_entity.IpAddresses = []
            host_dict[row.DeviceName] = host_entity
        # add device ID
        if row.DeviceId not in host_entity.DeviceIds:
            host_entity.DeviceIds.append(row.DeviceId)
        # Add IP Addresses from dev info record
        ip_set = {row.PublicIP}
        # and from the devicenetworkinfo results for this device
        ip_set.update(
            result.associated_hosts[
                result.associated_hosts.DeviceId == row.DeviceId
            ].IPAddress
        )
        ip_entities = convert_to_ip_entities(list(ip_set), geo_lookup=geo_lookup)
        host_ips = {ip.Address for ip in host_entity.IpAddresses}
        host_entity.IpAddresses.extend(
            [
                ip_entity
                for ip_entity in ip_entities
                if ip_entity.Address not in host_ips
            ]
        )
    return list(host_dict.values())


_MDE_DEV_NETWORK_EVT = """
DeviceNetworkEvents
| where RemoteIP == "{src_ip}"
| extend Computer = DeviceName
| summarize Count=count(), FirstOperation=min(TimeGenerated),
LastOperation=max(TimeGenerated) by Computer, LocalIP, RemoteIP,
ActionType, InitiatingProcessAccountName, InitiatingProcessFolderPath,
InitiatingProcessSHA256
"""


@set_text(docs=_CELL_DOCS, key="get_device_network")
def _get_device_net_connections(qry_prov, src_ip, result):
    nb_data_wait("DeviceNetworkEvents")
    result.network_connections = qry_prov.exec_query(
        _MDE_DEV_NETWORK_EVT.format(src_ip=src_ip)
    )
    if df_has_data(result.network_connections):
        _display_df_summary(result.network_connections, "Defender network connections")
        nb_display(
            result.network_connections[["Computer", "LocalIP"]].drop_duplicates()
        )
    else:
        nb_markdown("No events from Device network connections")


_WIN_HOSTS_LOGONS = """
SecurityEvent
| where TimeGenerated >= datetime({start}) and TimeGenerated <= datetime({end})
| where EventID in (4624, 4625)
| where IpAddress == \'{ip_address}\'
| project-rename LogonTypeCode=LogonType
| project-rename LogonType=LogonTypeName
| summarize Count=count(), FirstOperation=min(TimeGenerated),
LastOperation=max(TimeGenerated) by Computer, Account, LogonType
"""
_LX_HOST_LOGONS_ADD = """
| project-rename Account=User
| summarize Count=count(), FirstOperation=min(TimeGenerated),
LastOperation=max(TimeGenerated) by Computer, Account, LogonType
"""


@set_text(docs=_CELL_DOCS, key="get_host_logons")
def _get_host_logons(qry_prov, src_ip, result, timespan):
    """Get the hosts with logons for `src_ip`."""
    nb_data_wait("Related hosts")
    win_hosts_df = qry_prov.exec_query(
        _WIN_HOSTS_LOGONS.format(
            start=timespan.start,
            end=timespan.end,
            ip_address=src_ip,
        )
    )
    lx_hosts_df = qry_prov.LinuxSyslog.list_logons_for_source_ip(
        start=timespan.start,
        end=timespan.end,
        ip_address=src_ip,
        add_query_items=_LX_HOST_LOGONS_ADD,
    )
    combined_results = []
    if df_has_data(win_hosts_df):
        combined_results.append(win_hosts_df.assign(Type="Windows"))
    if df_has_data(lx_hosts_df):
        combined_results.append(lx_hosts_df.assign(Type="Linux"))
    if combined_results:
        result.host_logons = pd.concat(combined_results, ignore_index=True)[
            [
                "Computer",
                "Account",
                "Type",
                "LogonType",
                "Count",
                "FirstOperation",
                "LastOperation",
            ]
        ]
        nb_display(
            result.host_logons[
                ["Computer", "Type", "LogonType", "Count"]
            ].drop_duplicates()
        )
        return
    nb_markdown("No results for host logons")


@set_text(docs=_CELL_DOCS, key="get_related_accounts")
def _get_related_accounts(result):
    """Get the related accounts for `src_ip`."""
    nb_data_wait("Related accounts")
    combined_results = []
    if df_has_data(result.host_logons):
        combined_results.append(result.host_logons)
    if df_has_data(result.azure_activity_summary):
        combined_results.append(
            result.azure_activity_summary.reset_index().rename(
                columns={"UserPrincipalName": "Account", "OperationCount": "Count"}
            )[["Account", "Type", "Status", "Count", "FirstOperation", "LastOperation"]]
        )
    if df_has_data(result.office_activity):
        combined_results.append(
            result.office_activity.reset_index().rename(
                columns={
                    "OfficeWorkload": "Type",
                    "OperationCount": "Count",
                }
            )[["Account", "Type", "Count", "FirstOperation", "LastOperation"]]
        )

    if combined_results:
        result.related_accounts = pd.concat(combined_results, ignore_index=True)
        nb_display(
            result.related_accounts[["Account", "Type", "Count"]].drop_duplicates()
        )
        return
    nb_markdown("No results for related accounts")


def _populate_host_entity(result, geo_lookup=None):
    """Populate host entity and IP address details."""
    result.host_entities.append(
        populate_host_entity(
            heartbeat_df=result.heartbeat,
            az_net_df=result.az_network_if,
            vmcomputer_df=result.vmcomputer,
            geo_lookup=geo_lookup,
            host_entity=result.host_entities[0] if result.host_entities else None,
        )
    )


# %%
# Public IP functions
def _get_whois(src_ip, result):
    """Get WhoIs data and split out networks."""
    whois_result = get_whois_info(src_ip)
    if hasattr(whois_result, "properties"):
        whois_dict = whois_result.properties
    else:
        whois_dict = whois_result[1]
    result.whois = pd.DataFrame(whois_dict)
    result.whois_nets = pd.DataFrame(whois_dict.get("nets", []))
    if df_has_data(result.whois):
        nb_markdown("Whois data retrieved")


def _get_geoip_data(geo_lookup, src_ip, result):
    if result.ip_entity:
        geo_list, ip_list = geo_lookup.lookup_ip(ip_entity=result.ip_entity)
    else:
        geo_list, ip_list = geo_lookup.lookup_ip(src_ip)
    result.geoip = geo_list[0] if geo_list else None
    if isinstance(result.geoip, str):
        try:
            result.geoip = json.loads(result.geoip)
        except JSONDecodeError:
            pass
    result.ip_entity = ip_list[0] if ip_list else None
    if result.ip_entity and hasattr(result.ip_entity, "Location"):
        result.location = result.ip_entity.Location
        nb_markdown("GeoLocation data retrieved")
        nb_display(result.ip_entity.Location)


def _get_ti_data(ti_lookup, src_ip, result):
    nb_data_wait("Threat Intel")
    if not ti_lookup:
        return
    ti_results = ti_lookup.lookup_ioc(src_ip)
    result.ti_results = ti_lookup.result_to_df(ti_results)
    warn_ti_res = len(result.ti_results.query("Severity > 0"))
    if warn_ti_res:
        nb_markdown(f"{warn_ti_res} TI result(s) of severity 'warning' or above found.")
        nb_display(result.ti_results)
        nb_markdown("Use `browse_ti_results()` to view details.")


def _get_passv_dns(ti_lookup, src_ip, result):
    nb_data_wait("Passive DNS")
    if not ti_lookup:
        return
    try:
        ip_class = ip_address(src_ip)
    except ValueError:
        ip_class = None
    if not hasattr(ti_lookup, "available_query_types"):
        return
    if (
        isinstance(ip_class, IPv6Address)
        and "ipv6-passivedns" not in ti_lookup.available_query_types
    ):
        return
    if (
        isinstance(ip_class, IPv4Address)
        and "ipv4-passivedns" not in ti_lookup.available_query_types
    ):
        return
    with suppress(MsticpyException):
        passv_dns = ti_lookup.lookup_ioc(
            src_ip,
            ioc_type="ipv4" if isinstance(ip_class, IPv4Address) else "ipv6",
            ioc_query_type="passivedns",
        )
        result.passive_dns = ti_lookup.result_to_df(passv_dns)
    if result.passive_dns is not None and not result.passive_dns.empty:
        nb_markdown(f"{len(result.passive_dns)} Passive DNS results found.")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Threat Intelligence notebooklet feature support."""
from typing import Any, Optional, Tuple

import numpy as np
import pandas as pd

from .._version import VERSION
from ..common import nb_markdown

__version__ = VERSION
__author__ = "Pete Bryan"


def get_ti_results(
    ti_lookup,
    data: pd.DataFrame,
    col: str,
) -> Tuple[Any, Optional[pd.DataFrame]]:
    """
    Lookup Threat Intel.

    Parameters
    ----------
    ti_lookup : TILookup
        TI Lookup provider
    data : pd.DataFrame
        Input data frame
    col : List
        Name of Ip address column

    Returns
    -------
    pd.DataFrame
        DataFrame with TI results for IPs

    """
    ti_merged_df = None
    iocs = data[col].dropna().unique()
    nb_markdown(f"Querying TI for {len(iocs)} indicators...")
    ti_results = ti_lookup.lookup_iocs(data=data, ioc_col=col)
    if isinstance(ti_results, pd.DataFrame) and not ti_results.empty:
        ti_results = ti_results[ti_results["Severity"] > 0]
        ti_merged_df = data.merge(ti_results, how="inner", left_on=col, right_on="Ioc")
    return ti_results, ti_merged_df


def extract_iocs(
    data: pd.DataFrame, col: str, b64_extract: bool = False
) -> pd.DataFrame:
    """
    Extract IoCs from a dataframe.

    Parameters
    ----------
    data : pd.DataFrame
        The DataFrame to extract IoCs from
    col : str
        The column to extract IoCs from
    b64_extract : bool, optional
        wheter to attempt to B64 decode the column before extracting, by default False

    Returns
    -------
    pd.DataFrame
        DataFrame with extracted IoCs in the IoC column

    """
    if b64_extract:
        b64_extracted_data = data.mp_b64.extract(column=col)
        if not b64_extracted_data.empty:
            b64_extracted = pd.merge(
                data, b64_extracted_data[["decoded_string", col]], how="left", on=col
            )
            b64_extracted["decoded_string"] = b64_extracted["decoded_string"].astype(
                "str"
            )
            b64_iocs = b64_extracted.mp_ioc.extract(columns=["decoded_string"])
            b64_iocs["SourceIndex"] = pd.to_numeric(b64_iocs["SourceIndex"])
            data = pd.merge(
                left=data,
                right=b64_iocs,
                how="outer",
                left_index=True,
                right_on="SourceIndex",
            )

    iocs = data.mp_ioc.extract(columns=[col])
    data = pd.merge(
        left=data,
        right=iocs,
        how="outer",
        left_index=True,
        right_on="SourceIndex",
    )

    if "Observable_x" in data.columns:
        data["IoC"] = np.where(
            data["Observable_x"].isna(),
            data["Observable_y"],
            data["Observable_x"],
        )
        data["IoCType"] = np.where(
            data["IoCType_x"].isna(),
            data["IoCType_y"],
            data["IoCType_x"],
        )
        data["IoC"] = data["IoC"].astype("str")
    else:
        data["IoC"] = data["Observable"].astype("str")
    data["IoCType"] = data["IoCType"].astype("str")
    return data



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Notebooklet for Windows Security Events."""
import os
import pkgutil
from io import StringIO
from typing import Any, Dict, Iterable, Optional, Union

import numpy as np
import pandas as pd
from bokeh.models import LayoutDOM
from defusedxml import ElementTree
from defusedxml.ElementTree import ParseError
from IPython.display import display
from msticpy.common.timespan import TimeSpan

try:
    from msticpy.vis.timeline import display_timeline
except ImportError:
    # Fall back to msticpy locations prior to v2.0.0
    from msticpy.nbtools.nbdisplay import display_timeline

from .... import nb_metadata
from ...._version import VERSION
from ....common import (
    MsticnbMissingParameterError,
    nb_data_wait,
    nb_markdown,
    nb_warn,
    set_text,
)
from ....notebooklet import NBMetadata, Notebooklet, NotebookletResult

__version__ = VERSION
__author__ = "Ian Hellen"


_CLS_METADATA: NBMetadata
_CELL_DOCS: Dict[str, Any]
_CLS_METADATA, _CELL_DOCS = nb_metadata.read_mod_metadata(__file__, __name__)


# pylint: disable=too-few-public-methods
class WinHostEventsResult(NotebookletResult):
    """
    Windows Host Security Events Results.

    Attributes
    ----------
    all_events : pd.DataFrame
        DataFrame of all raw events retrieved.
    event_pivot : pd.DataFrame
        DataFrame that is a pivot table of event ID
        vs. Account
    account_events : pd.DataFrame
        DataFrame containing a subset of account management
        events such as account and group modification.
    acct_pivot : pd.DataFrame
        DataFrame that is a pivot table of event ID
        vs. Account of account management events
    account_timeline : Optional[LayoutDOM]
        Bokeh plot figure or Layout showing the account events on an
        interactive timeline.
    expanded_events : pd.DataFrame
        If `expand_events` option is specified, this will contain
        the parsed/expanded EventData as individual columns.

    """

    def __init__(
        self,
        description: Optional[str] = None,
        timespan: Optional[TimeSpan] = None,
        notebooklet: Optional["Notebooklet"] = None,
    ):
        """
        Create new Notebooklet result instance.

        Parameters
        ----------
        description : Optional[str], optional
            Result description, by default None
        timespan : Optional[TimeSpan], optional
            TimeSpan for the results, by default None
        notebooklet : Optional[, optional
            Originating notebooklet, by default None

        """
        super().__init__(description, timespan, notebooklet)
        self.description: str = "Windows Host Security Events"
        self.all_events: Optional[pd.DataFrame] = None
        self.event_pivot: Optional[pd.DataFrame] = None
        self.account_events: Optional[pd.DataFrame] = None
        self.account_pivot: Optional[pd.DataFrame] = None
        self.account_timeline: Optional[LayoutDOM] = None
        self.expanded_events: Optional[pd.DataFrame] = None


class WinHostEvents(Notebooklet):
    """
    Windows host Security Events Notebooklet class.

    Queries and displays Windows Security Events including:

    - All security events summary
    - Extracting and displaying account management events
    - Account management event timeline
    - Optionally parsing packed event data into DataFrame columns

    Process (4688) and Account Logon (4624, 4625) are not included
    in the event types processed by this module.

    """

    metadata = _CLS_METADATA
    __doc__ = nb_metadata.update_class_doc(__doc__, metadata)
    _cell_docs = _CELL_DOCS

    @set_text(docs=_CELL_DOCS, key="run")
    def run(
        self,
        value: Any = None,
        data: Optional[pd.DataFrame] = None,
        timespan: Optional[TimeSpan] = None,
        options: Optional[Iterable[str]] = None,
        **kwargs,
    ) -> WinHostEventsResult:
        """
        Return Windows Security Event summary.

        Parameters
        ----------
        value : str
            Host name
        data : Optional[pd.DataFrame], optional
            Not used, by default None
        timespan : TimeSpan
            Timespan over which operations such as queries will be
            performed, by default None.
            This can be a TimeStamp object or another object that
            has valid `start`, `end`, or `period` attributes.
        options : Optional[Iterable[str]], optional
            List of options to use, by default None.
            A value of None means use default options.
            Options prefixed with "+" will be added to the default options.
            To see the list of available options type `help(cls)` where
            "cls" is the notebooklet class or an instance of this class.

        Other Parameters
        ----------------
        start : Union[datetime, datelike-string]
            Alternative to specifying timespan parameter.
        end : Union[datetime, datelike-string]
            Alternative to specifying timespan parameter.

        Returns
        -------
        HostSummaryResult
            Result object with attributes for each result type.

        Raises
        ------
        MsticnbMissingParameterError
            If required parameters are missing

        """
        super().run(
            value=value, data=data, timespan=timespan, options=options, **kwargs
        )

        if not value:
            raise MsticnbMissingParameterError("value")
        if not timespan:
            raise MsticnbMissingParameterError("timespan.")

        result = WinHostEventsResult(
            notebooklet=self, description=self.metadata.description, timespan=timespan
        )

        all_events_df, event_pivot_df = _get_win_security_events(
            self.query_provider, host_name=value, timespan=self.timespan
        )
        result.all_events = all_events_df
        result.event_pivot = event_pivot_df

        if "event_pivot" in self.options:
            _display_event_pivot(event_pivot=event_pivot_df)

        if "acct_events" in self.options:
            result.account_events = _extract_acct_mgmt_events(event_data=all_events_df)
            result.account_pivot = _create_acct_event_pivot(
                account_event_data=result.account_events
            )
            if result.account_pivot is not None:
                _display_acct_event_pivot(event_pivot_df=result.account_pivot)
                result.account_timeline = _display_acct_mgmt_timeline(
                    acct_event_data=result.account_events
                )

        if "expand_events" in self.options:
            result.expanded_events = _parse_eventdata(all_events_df)

        nb_markdown("To unpack eventdata from selected events use expand_events()")
        self._last_result = result  # pylint: disable=attribute-defined-outside-init
        return self._last_result

    def expand_events(
        self, event_ids: Optional[Union[int, Iterable[int]]] = None
    ) -> Optional[pd.DataFrame]:
        """
        Expand `EventData` for `event_ids` into separate columns.

        Parameters
        ----------
        event_ids : Optional[Union[int, Iterable[int]]], optional
            Single or interable of event IDs (ints).
            If no event_ids are specified all events will be expanded.

        Returns
        -------
        pd.DataFrame
            Results with expanded columns.

        Notes
        -----
        For a specific event ID you can expand the EventProperties values
        into their own columns using this function.
        You can do this for the whole data set but it will time-consuming
        and result in a lot of sparse columns in the output data frame.

        """
        if (
            not self._last_result or self._last_result.all_events is None
        ):  # type: ignore
            print(
                "Please use 'run()' to fetch the data before using this method.",
                "\nThen call 'expand_events()'",
            )
            return None
        return _parse_eventdata(
            event_data=self._last_result.all_events,  # type: ignore
            event_ids=event_ids,
        )


# %%
# Get Windows Security Events
def _get_win_security_events(qry_prov, host_name, timespan):
    nb_data_wait("SecurityEvent")

    all_events_df = qry_prov.WindowsSecurity.list_host_events(
        timespan,
        host_name=host_name,
        add_query_items="| where EventID != 4688 and EventID != 4624",
    )

    # Create a pivot of Event vs. Account
    win_events_acc = all_events_df[["Account", "Activity", "TimeGenerated"]].copy()
    win_events_acc = win_events_acc.replace("-\\-", "No Account").replace(
        {"Account": ""}, value="No Account"
    )
    win_events_acc["Account"] = win_events_acc.apply(
        lambda x: x.Account.split("\\")[-1], axis=1
    )
    event_pivot_df = (
        pd.pivot_table(
            win_events_acc,
            values="TimeGenerated",
            index=["Activity"],
            columns=["Account"],
            aggfunc="count",
        )
        .fillna(0)
        .reset_index()
    )
    return all_events_df, event_pivot_df


@set_text(docs=_CELL_DOCS, key="display_event_pivot")
def _display_event_pivot(event_pivot):
    if pd.__version__ < "2.1.0":
        styled_data = (
            event_pivot.style.applymap(lambda x: "color: white" if x == 0 else "")
            .applymap(
                lambda x: (
                    "background-color: lightblue"
                    if not isinstance(x, str) and x > 0
                    else ""
                )
            )
            .set_properties(
                subset=["Activity"], **{"width": "400px", "text-align": "left"}
            )
            .highlight_max(axis=1)
            .hide(axis="index")
        )
    else:
        styled_data = (
            event_pivot.style.map(lambda x: "color: white" if x == 0 else "")
            .map(
                lambda x: (
                    "background-color: lightblue"
                    if not isinstance(x, str) and x > 0
                    else ""
                )
            )
            .set_properties(
                subset=["Activity"], **{"width": "400px", "text-align": "left"}
            )
            .highlight_max(axis=1)
            .hide(axis="index")
        )
    display(styled_data)


# %%
# Extract event details from events
SCHEMA = "http://schemas.microsoft.com/win/2004/08/events/event"


def _parse_event_data_row(row):
    try:
        xdoc = ElementTree.fromstring(row.EventData)
        col_dict = {
            elem.attrib["Name"]: elem.text for elem in xdoc.findall(f"{{{SCHEMA}}}Data")
        }
        reassigned = set()
        for key, val in col_dict.items():
            if key in row and not row[key]:
                row[key] = val
                reassigned.add(key)
        if reassigned:
            for key in reassigned:
                col_dict.pop(key)
        return col_dict
    except (ParseError, TypeError):
        return None


def _expand_event_properties(input_df):
    # For a specific event ID you can explode the EventProperties values
    # into their own columns using this function. You can do this for
    # the whole data set but it will result
    # in a lot of sparse columns in the output data frame.
    exp_df = input_df.apply(lambda x: pd.Series(x.EventProperties), axis=1)
    with pd.option_context("future.no_silent_downcasting", True):
        return (
            exp_df.drop(set(input_df.columns).intersection(exp_df.columns), axis=1)
            .merge(
                input_df.drop("EventProperties", axis=1),
                how="inner",
                left_index=True,
                right_index=True,
            )
            .replace(
                r"^\s*$", np.nan, regex=True
            )  # these 3 lines get rid of blank columns
            .dropna(axis=1, how="all")
            .infer_objects(copy=False)
            .fillna("")
        )


@set_text(docs=_CELL_DOCS, key="parse_eventdata")
def _parse_eventdata(event_data, event_ids: Optional[Union[int, Iterable[int]]] = None):
    if event_ids:
        if isinstance(event_ids, int):
            event_ids = [event_ids]
        src_event_data = event_data[event_data["EventID"].isin(event_ids)].copy()
    else:
        src_event_data = event_data.copy()

    if src_event_data.empty:
        nb_warn(f"No events matching {event_ids}")
        return None
    # Parse event properties into a dictionary
    nb_markdown("Parsing event data...")
    src_event_data["EventProperties"] = src_event_data.apply(
        _parse_event_data_row, axis=1
    )
    return _expand_event_properties(src_event_data)


# %%
# Account management events
def _extract_acct_mgmt_events(event_data):
    # Get a full list of Windows Security Events

    w_evt = pkgutil.get_data("msticpy", f"resources{os.sep}WinSecurityEvent.json")

    win_event_df = pd.read_json(StringIO(w_evt.decode("utf-8")))

    # Create criteria for events that we're interested in
    acct_sel = win_event_df["subcategory"] == "User Account Management"
    group_sel = win_event_df["subcategory"] == "Security Group Management"
    schtask_sel = (win_event_df["subcategory"] == "Other Object Access Events") & (
        win_event_df["description"].str.contains("scheduled task")
    )

    event_list = win_event_df[acct_sel | group_sel | schtask_sel]["event_id"].to_list()
    # Add Service install event
    event_list.append(7045)
    return event_data[event_data["EventID"].isin(event_list)]


def _create_acct_event_pivot(account_event_data):
    # Create a pivot of Event vs. Account
    if account_event_data.empty:
        return None
    win_events_acc = account_event_data[["Account", "Activity", "TimeGenerated"]].copy()
    win_events_acc = win_events_acc.replace("-\\-", "No Account").replace(
        {"Account": ""}, value="No Account"
    )
    win_events_acc["Account"] = win_events_acc.apply(
        lambda x: x.Account.split("\\")[-1], axis=1
    )

    return (
        pd.pivot_table(
            win_events_acc,
            values="TimeGenerated",
            index=["Activity"],
            columns=["Account"],
            aggfunc="count",
        )
        .fillna(0)
        .reset_index()
    )


@set_text(docs=_CELL_DOCS, key="display_acct_event_pivot")
def _display_acct_event_pivot(event_pivot_df):
    if pd.__version__ < "2.1.0":
        styled_data = (
            event_pivot_df.style.applymap(lambda x: "color: white" if x == 0 else "")
            .applymap(
                lambda x: (
                    "background-color: lightblue"
                    if not isinstance(x, str) and x > 0
                    else ""
                )
            )
            .set_properties(
                subset=["Activity"], **{"width": "400px", "text-align": "left"}
            )
            .highlight_max(axis=1)
            .hide(axis="index")
        )
    else:
        styled_data = (
            event_pivot_df.style.map(lambda x: "color: white" if x == 0 else "")
            .map(
                lambda x: (
                    "background-color: lightblue"
                    if not isinstance(x, str) and x > 0
                    else ""
                )
            )
            .set_properties(
                subset=["Activity"], **{"width": "400px", "text-align": "left"}
            )
            .highlight_max(axis=1)
            .hide(axis="index")
        )
    display(styled_data)


@set_text(docs=_CELL_DOCS, key="display_acct_mgmt_timeline")
def _display_acct_mgmt_timeline(acct_event_data):
    # Plot events on a timeline
    return display_timeline(
        data=acct_event_data,
        group_by="EventID",
        source_columns=["Activity", "Account"],
        legend="right",
    )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Entity Helper functions."""
from typing import Dict, List, Union

import pandas as pd

from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


def extract_entities(
    data: pd.DataFrame, cols: Union[str, List[str]]
) -> Dict[str, List[str]]:
    """
    Extract items from a column (strings or lists).

    Parameters
    ----------
    data : pd.DataFrame
        DataFrame to parse
    cols : Union[str, List[str]]
        Columns to use for input

    Returns
    -------
    Dict[str, List[str]]
        Dictionary of (column: result_list)

    """
    if not isinstance(cols, list):
        cols = [cols]

    val_results = {}
    for col in cols:
        ent_vals = list(data[col].values)
        test_val = data[col].iloc[0]
        if isinstance(test_val, list):
            ent_vals = list({ent for ent_list in ent_vals for ent in ent_list})
        val_results[col] = ent_vals

    return val_results



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""read_modules - handles reading notebooklets modules."""
import importlib
import inspect
import sys
from collections import namedtuple
from functools import partial
from operator import itemgetter
from pathlib import Path
from typing import Dict, Iterable, List, Tuple, Type, Union
from warnings import warn

from . import nb
from ._version import VERSION
from .class_doc import get_class_doc
from .common import MsticnbError, NBContainer, nb_debug
from .notebooklet import Notebooklet

__version__ = VERSION
__author__ = "Ian Hellen"

nblts: NBContainer = NBContainer()
nb_index: Dict[str, Type[Notebooklet]] = {}


def discover_modules(nb_path: Union[str, Iterable[str], None] = None) -> NBContainer:
    """
    Discover notebooks modules.

    Parameters
    ----------
    nb_path : Union[str, Iterable[str]], optional
        Additional path to search for notebooklets, by default None

    Returns
    -------
    NBContainer
        Container of notebooklets. This is structured
        as a tree mirroring the source folder names.

    """
    pkg_folder = Path(__file__).parent
    nb_folder = pkg_folder / "nb"
    _import_from_folder(nb_folder, pkg_folder)

    # Import from user-defined folders
    if not nb_path:
        return nblts
    if isinstance(nb_path, str):
        nb_path = [nb_path]
    if isinstance(nb_path, list):
        for path_item in nb_path:
            # For custom notebooklets, we need to add the path
            # to the source folder to sys.path so that import can
            # find and import them.
            cust_nb_path = Path(path_item).resolve()
            sys.path.append(str(cust_nb_path.parent))
            _import_from_folder(cust_nb_path, cust_nb_path)
    return nblts


def _import_from_folder(nb_folder: Path, pkg_folder: Path):
    """Search folder and import any notebooklets found."""
    if not nb_folder.is_dir():
        raise MsticnbError(f"Notebooklet folder {nb_folder} not found.")

    # if we are importing notebooklets from outside this package, we
    # specify customer_container to put them in their own subtree.
    custom_container = pkg_folder.stem if pkg_folder.stem != __package__ else ""

    # Iterate through folder and all subfolders
    folders = [
        nb_folder,
        *(f for f in nb_folder.glob("./**") if f.is_dir() and f != nb_folder),
    ]
    for folder in folders:
        rel_folder_parts = folder.relative_to(nb_folder).parts
        # skip hidden folder paths with . or _ prefix
        if any(f for f in rel_folder_parts if f.startswith(".") or f.startswith("_")):
            continue

        # Get any notebooklets from the files in the folder
        nb_classes = _find_cls_modules(folder, pkg_folder)
        if not nb_classes:
            continue
        # Get the container to add these classes to the container
        # tree
        cur_container = _get_container(custom_container, rel_folder_parts)
        for cls_name, nb_class in nb_classes.items():
            setattr(cur_container, cls_name, nb_class)
            cls_index = "nblts." + ".".join(list(rel_folder_parts) + [cls_name])
            nb_index[cls_index] = nb_class


def _find_cls_modules(folder: Path, pkg_folder: Path) -> Dict[str, Type[Notebooklet]]:
    """
    Import .py files in `folder` and return any Notebooklet classes found.

    Parameters
    ----------
    folder : Path
        The folder to search
    pkg_folder : Path
        The root path for the package

    Returns
    -------
    Dict[str, type(Notebooklet)]
        Notebooklets classes (name, class)

    """
    found_classes = {}
    pkg_root = pkg_folder.resolve().parent
    try:
        relative_path = folder.relative_to(pkg_root)
        # This may fail if environment messes around with package paths
        # - this happens in Spark/Synapse
    except ValueError:
        relative_path = _get_pkg_relative_folder(folder)
    for item in folder.glob("*.py"):
        if item.name.startswith("_"):
            continue
        if item.is_file():
            # Create full package path for item
            mod_name = ".".join(list(relative_path.parts) + [item.stem])
            nb_debug("module to import", item, mod_name)
            # Try to import the module into this package
            try:
                imp_module = importlib.import_module(mod_name, package=nb.__package__)
            except ImportError as err:
                warn(f"Import failed for {item}.\n {err}")
                nb_debug("import failed", item, err)
                continue
            # extract the classes in the module and look for any classes
            # derived from Notebooklet
            mod_classes = inspect.getmembers(imp_module, inspect.isclass)
            for cls_name, mod_class in mod_classes:
                if issubclass(mod_class, Notebooklet) and mod_class != Notebooklet:
                    nb_debug("imported", cls_name)
                    # We need to store the path of the parent module in the class
                    # - this makes it easier to retrieve when we need it for
                    # reading metadata and generating the class docs.
                    mod_class.module_path = str(item)
                    # create a function (pointer) in the class that will
                    # build and return our extended class documentation
                    setattr(
                        mod_class, "_get_doc", partial(get_class_doc, doc_cls=mod_class)
                    )
                    found_classes[cls_name] = mod_class
    return found_classes


def _get_container(custom_cont: str, path_parts: Tuple[str, ...]) -> NBContainer:
    """
    Return container corresponding to path_parts.

    This will search through existing container tree for
    a match for path_parts (optionally prefixed with `custom_cont`).
    If the container (and any required parent containers) does not
    exist it will be created (any required parents to the container
    will also be created).

    Parameters
    ----------
    custom_cont : str
        An optional prefix element for the path to be
        searched
    path_parts : Tuple[str, ...]
        A tuple of path elements for the tree position

    Returns
    -------
    NBContainer
        The container item in the tree. New one created if it
        does not exist.

    """
    cur_container = nblts
    path_elems = list(path_parts)
    if custom_cont:
        path_elems = [custom_cont] + list(path_elems)
    for path_item in path_elems:
        child_item = getattr(cur_container, path_item, None)
        if not child_item:
            child_item = NBContainer()
            setattr(cur_container, path_item, child_item)
        cur_container = child_item
    return cur_container


FindResult = namedtuple("FindResult", "full_match match_count, name, nb_class")


def find(keywords: str, full_match=True) -> List[Tuple[str, Notebooklet]]:
    """
    Search for Notebooklets matching key words.

    Parameters
    ----------
    keywords : str
        Space or comma-separated words to search for.
        Terms can be regular expressions.
    full_match : bool
        If True only return full matches, default is True.
        If False it will return partial matches.

    Returns
    -------
    List[Tuple[str, Notebooklet]]
        List of matches sorted by closest match

    Notes
    -----
    Search terms are treated as regular expressions, so any regular
    expression reserved characters will be treated as part of the
    regex pattern.

    """
    matches = []
    for name, nb_class in nblts.iter_classes():
        all_match, match_count = nb_class.match_terms(keywords)
        if all_match or (match_count and not full_match):
            matches.append(FindResult(all_match, match_count, name, nb_class))

    # return list sorted by full_match, then match count, highest to lowest
    results = sorted(matches, key=itemgetter(0, 1), reverse=True)
    return [(result[2], result[3]) for result in results]


def _get_pkg_relative_folder(folder: Path):
    """Return relative path based on location of 'msticnb' in the folder path."""
    relative_parts: List[str] = []
    for folder_part in folder.parts:
        if folder_part == "msticnb" or relative_parts:
            relative_parts.append(folder_part)
    return Path(*relative_parts)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""read_modules test class."""
from pathlib import Path

import pytest_check as check

from msticnb.read_modules import Notebooklet, discover_modules, find, nb_index, nblts

from .unit_test_lib import TEST_DATA_PATH


def test_read_modules():
    """Test method."""
    nbklts = discover_modules()
    check.greater_equal(len(list(nbklts.iter_classes())), 4)

    # pylint: disable=no-member
    match, m_count = nblts.azsent.host.HostSummary.match_terms("host, linux, azure")
    check.is_true(match)
    check.equal(m_count, 3)

    for key, value in nbklts.iter_classes():
        check.is_instance(key, str)
        check.is_true(issubclass(value, Notebooklet))

    find_res = find("host windows azure")
    check.greater(len(find_res), 0)
    not_found = find("monkey stew")
    check.equal(len(not_found), 0)


def test_read_custom_path():
    """Test method."""
    cust_nb_path = Path(TEST_DATA_PATH).parent / "custom_nb"
    nbklts = discover_modules(nb_path=str(cust_nb_path))
    check.greater_equal(len(list(nbklts.iter_classes())), 5)

    # pylint: disable=no-member
    match, m_count = nblts.custom_nb.host.CustomNB.match_terms("Custom")
    check.is_true(match)
    check.equal(m_count, 1)

    for key, value in nbklts.iter_classes():
        check.is_instance(key, str)
        check.is_true(issubclass(value, Notebooklet))

    find_res = find("banana")
    check.equal(len(find_res), 1)
    find_res = find("<<Test Marker>>")
    check.equal(len(find_res), 1)
    check.equal(find_res[0][0], "CustomNB")
    check.is_in("nblts.host.CustomNB", nb_index)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Jupyter Browser for Notebooklets."""

import inspect
import re
from datetime import datetime, timedelta

import ipywidgets as widgets
from IPython import get_ipython
from IPython.display import display
from markdown import markdown

from ._version import VERSION
from .common import MsticnbError
from .read_modules import nb_index, nblts

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=too-few-public-methods
class NBBrowser:
    """Interactive browser/viewer for Notebooklets."""

    def __init__(self):
        """Initialize and Display Notebooklet Browser."""
        # Define Controls
        nb_list = list(nblts.iter_classes())

        if not nb_list:
            raise MsticnbError("No notebooklets are loaded.")
        self.selected = nb_list[0][1]
        self.nb_select = widgets.Select(options=nb_list)
        nb_doc_layout = widgets.Layout(
            width="60%",
            border="1px solid lightgray",
            margin="0px 0px 0px 10px",
        )
        self.nb_doc = widgets.HTML(layout=nb_doc_layout)
        nb_details_layout = widgets.Layout(width="100%")
        self.nb_code = widgets.HTML(layout=nb_details_layout)
        self.nb_details = widgets.HTML(layout=nb_details_layout)
        self.nb_run_doc = widgets.HTML(layout=nb_details_layout)

        self.nb_select.observe(self._update_nbdetails, names="value")

        btn_insert_class = widgets.Button(description="Insert code example")
        btn_insert_class.on_click(self._insert_code)

        # Accordian Group
        nb_det_accdn = widgets.Accordion(
            children=[self.nb_details, self.nb_run_doc, self.nb_code]
        )
        nb_det_accdn.set_title(0, "Full class details")
        nb_det_accdn.set_title(1, "run() method documentation")
        nb_det_accdn.set_title(2, "Sample code")
        nb_det_accdn.selected_index = None

        # Update display based on first item in the list
        self._update_nbdetails({"new": nb_list[0][1]})

        # Widget layout
        vbox = widgets.VBox([self.nb_select, btn_insert_class])
        top_hbox = widgets.HBox([vbox, self.nb_doc])
        self._all_controls = widgets.VBox([top_hbox, nb_det_accdn])
        display(self._all_controls)

    def display(self):
        """Display the widget."""
        display(self._all_controls)

    @staticmethod
    def _get_class_index(tgt_class):
        for name, val in nb_index.items():
            if tgt_class is val:
                return name
        return None

    @staticmethod
    def _pyvar_case(in_name):
        return "_".join(
            part.lower() for part in re.split(r"([A-Z][a-z]*)", in_name) if part
        )

    def _update_nbdetails(self, change):
        nb_cls = change["new"]
        self.selected = nb_cls
        nb_html = f"<h1>{nb_cls.__name__}</h1>"
        nb_html += markdown(inspect.cleandoc(nb_cls.__doc__))
        self.nb_doc.value = nb_html
        self._populate_docs(self.nb_select.value)

    def _populate_docs(self, nb_cls):
        nb_html = nb_cls.get_help()
        self.nb_details.value = nb_html
        nb_run_html = f"<h1>{nb_cls.__name__}.run()</h1>"
        nb_run_html += markdown(nb_cls.run.__doc__)
        self.nb_run_doc.value = nb_run_html
        self.nb_code.value = (
            f"<pre><code>{self._create_code_sample(nb_cls)}</code></pre>"
        )

    _CODE_TEMPLATE = """
    from msticpy.common.timespan import TimeSpan
    # Note you can populate TimeSpan with msticpy nbwidgets.QueryTime()
    # qt = QueryTime(autodisplay=True)
    # %% New cell
    # notebooklet.run(value="158.23.4.19", timespan=TimeSpan(qt))

    {nb_instance} = nb.{nb_cls}()
    timespan = TimeSpan(start="{start}")
    {nb_instance}_rslt = {nb_instance}.run(value="my_host", timespan=timespan)

    """

    def _create_code_sample(self, nb_cls):
        start = (datetime.utcnow() - timedelta(1)).strftime("%Y-%m-%d %H:%M")
        return self._CODE_TEMPLATE.format(
            nb_instance=self._pyvar_case(nb_cls.__name__),
            nb_cls=self._get_class_index(nb_cls),
            start=start,
        )

    def _insert_code(self, change):
        del change
        code = self._create_code_sample(self.nb_select.value)
        shell = get_ipython()
        if shell:
            shell.set_next_input(code)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Test module for nb_pivot."""
from pathlib import Path

import pytest
import pytest_check as check
from msticpy.datamodel import entities

try:
    from msticpy.init.pivot import Pivot
except ImportError:
    # Fall back to msticpy locations prior to v2.0.0
    from msticpy.datamodel.pivot import Pivot

from msticnb import data_providers, init, nblts
from msticnb.nb_pivot import add_pivot_funcs
from msticnb.notebooklet import NotebookletResult

from .unit_test_lib import TEST_DATA_PATH, GeoIPLiteMock

__author__ = "Ian Hellen"

# pylint: disable=redefined-outer-name

_EXPECTED_FUNCS = [
    (
        "Host",
        (
            "host_logons_summary",
            "host_summary",
            "win_host_events",
            "network_flow_summary",
        ),
        "test_host",
    ),
    ("Account", ("account_summary",), "test_acct"),
    ("IpAddress", ("network_flow_summary",), "11.1.2.3"),
]


@pytest.fixture
def _init_pivot(monkeypatch):
    init()
    test_data = str(Path(TEST_DATA_PATH).absolute())
    monkeypatch.setattr(data_providers, "GeoLiteLookup", GeoIPLiteMock)
    if "azuredata" in nblts.azsent.host.HostSummary.metadata.req_providers:
        nblts.azsent.host.HostSummary.metadata.req_providers.remove("azuredata")
    data_providers.init(
        query_provider="LocalData",
        providers=["geolitelookup"],
        LocalData_data_paths=[test_data],
        LocalData_query_paths=[test_data],
    )
    return Pivot()


@pytest.mark.parametrize("ent_name, funcs, test_val", _EXPECTED_FUNCS)
def test_add_pivot_funcs(_init_pivot, ent_name, funcs, test_val):
    """Test adding notebooklets to pivot."""
    del test_val
    add_pivot_funcs(_init_pivot)

    entity = getattr(entities, ent_name)
    container = getattr(entity, "nblt")
    for func_name in funcs:
        check.is_true(hasattr(container, func_name))


@pytest.mark.parametrize("ent_name, funcs, test_val", _EXPECTED_FUNCS)
def test_run_pivot_funcs(_init_pivot, ent_name, funcs, test_val):
    """Test running notebooklets run functions."""
    add_pivot_funcs(_init_pivot)

    entity = getattr(entities, ent_name)

    check.is_true(hasattr(entity, "nblt"))
    container = getattr(entity, "nblt")
    for _, p_func in container:
        if p_func.__name__ not in funcs:
            continue
        check.is_true(callable(p_func))
        result = p_func(value=test_val)
        test_result = result[0] if isinstance(result, list) else result
        check.is_true(isinstance(test_result, NotebookletResult))
        check.equal(test_result.timespan, _init_pivot.get_timespan())



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
msticnb Notebooklets main package.

To start using notebooklets:

.. code:: python
    >>> import msticnb as nb
    >>> # optionally give a query provider nb.init(query_provider=qry_prov)
    >>> nb.init()
    >>>
    >>> # Auto-complete tree of notebooklets
    >>> nb.nblts
    >>>
    >>> # List notebooklets
    >>> nb.nb_index
    >>>
    >>> # Use a notebooklet
    >>> host_summary = nb.nblts.azent.host.HostSummary()
    >>> host_summary.run();
    >>>
    >>> # help
    >>> help(host_summary)
    >>> print("Options:", host_summary.all_options())
    >>> print("Settings:", host_summary.get_settings())
    >>>
    >>> # find a notebooklet
    >>> nb.find("host linux azure")
    >>>
    >>> # Interactive notebook browser
    >>> nb.browse()

for more help see https://msticnb.readthedocs.org/

"""
import sys
from typing import Any, Dict, List, Optional, Union

from ._version import VERSION
from .data_providers import DataProviders, QueryProvider  # noqa:F401
from .data_providers import init as dp_init  # noqa:F401
from .nb_browser import NBBrowser  # noqa:F401
from .nb_pivot import add_pivot_funcs  # noqa:F401
from .options import get_opt, set_opt  # noqa:F401
from .read_modules import discover_modules, find, nb_index, nblts  # noqa:F401
from .template import create_template  # noqa:F401

__version__ = VERSION

# pylint: disable=invalid-name
browse = NBBrowser


def init(
    query_provider: Union[str, QueryProvider] = "MSSentinel",
    namespace: Optional[Dict[str, Any]] = None,
    providers: Optional[List[str]] = None,
    **kwargs,
):
    """
    Initialize notebooklets data providers and pivots.

    Parameters
    ----------
    query_provider : Union[str, QueryProvider], optional
        DataEnvironment name of the primary query provider,
        or an instance of an existing query provider,
        by default "MSSentinel"
    namespace : Optional[Dict[str, Any]], optional
        The global namespace - used to add pivot functions
    providers : Optional[List[str]], optional
        A list of other provider names to load

    Other Parameters
    ----------------
    kwargs :
        Optional keyword arguments to pass to DataProviders
        and Pivot initializers.

    Notes
    -----
    Use msticnb.DataProviders.list_providers() to get a list
    of accepted providers.

    """
    discover_modules()
    print(f"Notebooklets: {len(list(nblts.iter_classes()))} notebooklets loaded.")
    dp_init(query_provider=query_provider, providers=providers, **kwargs)
    if not namespace:
        # Try to get the globals namespace from top-level caller
        # pylint: disable=protected-access
        namespace = sys._getframe(1).f_globals
        # pylint: enable=protected-access
    add_pivot_funcs(namespace=namespace, **kwargs)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Notebooklet for Host Summary."""
from functools import lru_cache
from typing import Any, Dict, Iterable, Optional

import pandas as pd
from azure.common.exceptions import CloudError
from IPython.display import display

try:
    from msticpy import nbwidgets
    from msticpy.vis.timeline import display_timeline
except ImportError:
    # Fall back to msticpy locations prior to v2.0.0
    from msticpy.nbtools import nbwidgets
    from msticpy.nbtools.nbdisplay import display_timeline

from msticpy.common.timespan import TimeSpan
from msticpy.datamodel import entities

from ...._version import VERSION
from ....common import (
    MsticnbDataProviderError,
    MsticnbMissingParameterError,
    nb_data_wait,
    nb_markdown,
    nb_print,
    set_text,
)
from ....nb_metadata import read_mod_metadata, update_class_doc
from ....nblib.azsent.alert import browse_alerts
from ....nblib.azsent.host import get_aznet_topology, get_heartbeat, verify_host_name
from ....nblib.ti import extract_iocs, get_ti_results
from ....notebooklet import NBMetadata, Notebooklet, NotebookletResult

__version__ = VERSION
__author__ = "Ian Hellen"


_CLS_METADATA: NBMetadata
_CELL_DOCS: Dict[str, Any]
_CLS_METADATA, _CELL_DOCS = read_mod_metadata(__file__, __name__)


# pylint: disable=too-few-public-methods
class HostSummaryResult(NotebookletResult):
    """
    Host Details Results.

    Attributes
    ----------
    host_entity : msticpy.datamodel.entities.Host
        The host entity object contains data about the host
        such as name, environment, operating system version,
        IP addresses and Azure VM details. Depending on the
        type of host, not all of this data may be populated.
    related_alerts : pd.DataFrame
        Pandas DataFrame of any alerts recorded for the host
        within the query time span.
    related_bookmarks: pd.DataFrame
        Pandas DataFrame of any investigation bookmarks
        relating to the host.
    events: pd.DataFrame
        Pandas DataFrame of any high severity events
        from the selected host.

    """

    def __init__(
        self,
        description: Optional[str] = None,
        timespan: Optional[TimeSpan] = None,
        notebooklet: Optional["Notebooklet"] = None,
    ):
        """
        Create new Notebooklet result instance.

        Parameters
        ----------
        description : Optional[str], optional
            Result description, by default None
        timespan : Optional[TimeSpan], optional
            TimeSpan for the results, by default None
        notebooklet : Optional[], optional
            Originating notebooklet, by default None

        """
        super().__init__(description, timespan, notebooklet)
        self.host_entity: entities.Host = None
        self.related_alerts: Optional[pd.DataFrame] = None
        self.related_bookmarks: Optional[pd.DataFrame] = None
        self.summary: Optional[pd.DataFrame] = None
        self.scheduled_tasks: Optional[pd.DataFrame] = None
        self.account_actions: Optional[pd.DataFrame] = None
        self.notable_events: Optional[pd.DataFrame] = None
        self.processes: Optional[pd.DataFrame] = None
        self.process_ti: Optional[pd.DataFrame] = None


# pylint: disable=too-few-public-methods
class HostSummary(Notebooklet):
    """
    HostSummary Notebooklet class.

    Queries and displays information about a host including:

    - IP address assignment
    - Related alerts
    - Related hunting/investigation bookmarks
    - Azure subscription/resource data.

    """

    metadata = _CLS_METADATA
    __doc__ = update_class_doc(__doc__, metadata)
    _cell_docs = _CELL_DOCS

    def __init__(self, *args, **kwargs):
        """Initialize the Host Summary notebooklet."""
        super().__init__(*args, **kwargs)

    # pylint: disable=too-many-branches, too-many-statements
    def run(  # noqa:MC0001, C901
        self,
        value: Any = None,
        data: Optional[pd.DataFrame] = None,
        timespan: Optional[TimeSpan] = None,
        options: Optional[Iterable[str]] = None,
        **kwargs,
    ) -> HostSummaryResult:
        """
        Return host summary data.

        Parameters
        ----------
        value : str
            Host name
        data : Optional[pd.DataFrame], optional
            Not used, by default None
        timespan : TimeSpan
            Timespan over which operations such as queries will be
            performed, by default None.
            This can be a TimeStamp object or another object that
            has valid `start`, `end`, or `period` attributes.
        options : Optional[Iterable[str]], optional
            List of options to use, by default None
            A value of None means use default options.
            Options prefixed with "+" will be added to the default options.
            To see the list of available options type `help(cls)` where
            "cls" is the notebooklet class or an instance of this class.

        Other Parameters
        ----------------
        start : Union[datetime, datelike-string]
            Alternative to specifying timespan parameter.
        end : Union[datetime, datelike-string]
            Alternative to specifying timespan parameter.

        Returns
        -------
        HostSummaryResult
            Result object with attributes for each result type.

        Raises
        ------
        MsticnbMissingParameterError
            If required parameters are missing

        """
        super().run(
            value=value, data=data, timespan=timespan, options=options, **kwargs
        )

        if not value:
            raise MsticnbMissingParameterError("value")
        if not timespan:
            raise MsticnbMissingParameterError("timespan.")

        self.timespan = timespan

        # pylint: disable=attribute-defined-outside-init
        result = HostSummaryResult(
            notebooklet=self, description=self.metadata.description, timespan=timespan
        )

        host_verif = verify_host_name(self.query_provider, value, self.timespan)
        if host_verif.host_names:
            nb_markdown(f"Could not obtain unique host name from {value}. Aborting.")
            self._last_result = result
            return self._last_result
        if not host_verif.host_name:
            nb_markdown(
                f"Could not find event records for host {value}. "
                + "Results may be unreliable.",
                "orange",
            )
            self.host_name = value
        else:
            self.host_name = host_verif.host_name

        host_entity = entities.Host(HostName=self.host_name)
        if "heartbeat" in self.options:
            host_entity = get_heartbeat(self.query_provider, self.host_name)
        if "azure_net" in self.options:
            host_entity = host_entity or entities.Host(HostName=self.host_name)
            get_aznet_topology(
                self.query_provider, host_entity=host_entity, host_name=self.host_name
            )
        # If azure_details flag is set, an encrichment provider is given,
        # and the resource is an Azure host get resource details from Azure API

        if (
            "azure_api" in self.options
            and "azuredata" in self.data_providers.providers
            and ("Environment" in host_entity and host_entity.Environment == "Azure")
        ):
            azure_api = _azure_api_details(
                self.data_providers["azuredata"], host_entity
            )
            if azure_api:
                host_entity.AzureDetails["ResourceDetails"] = azure_api[
                    "resource_details"
                ]
                host_entity.AzureDetails["SubscriptionDetails"] = azure_api[
                    "sub_details"
                ]

        result.host_entity = host_entity

        if "alerts" in self.options:
            related_alerts = _get_related_alerts(
                self.query_provider, self.timespan, self.host_name
            )
            result.related_alerts = related_alerts

        if "bookmarks" in self.options:
            result.related_bookmarks = _get_related_bookmarks(
                self.query_provider, self.timespan, result.host_entity.HostName
            )

        if "scheduled_tasks" in self.options:
            result.scheduled_tasks = _get_scheduled_tasks(
                self.query_provider,
                self.timespan,
                result.host_entity.HostName,
                result.host_entity.OSFamily,
            )

        if "account_actions" in self.options:
            result.account_actions = _get_account_actions(
                self.query_provider,
                self.timespan,
                result.host_entity.HostName,
                result.host_entity.OSFamily,
            )

        if "notable_events" in self.options:
            result.notable_events = _get_notable_events(
                self.query_provider,
                self.timespan,
                result.host_entity.HostName,
                result.host_entity.OSFamily,
            )

        if "processes" in self.options:
            result.processes = _get_process_events(
                self.query_provider,
                self.timespan,
                result.host_entity.HostName,
                result.host_entity.OSFamily,
            )

        if (
            "process_ti" in self.options
            and isinstance(result.processes, pd.DataFrame)
            and not result.processes.empty
        ):
            cmd_column = (
                "CommandLine"
                if result.host_entity.OSFamily.name == "Windows"
                else "SyslogMessage"
            )
            if "tilookup" in self.data_providers.providers:
                ti_prov = self.data_providers.providers["tilookup"]
            else:
                raise MsticnbDataProviderError("No TI providers avaliable")
            result.process_ti = _process_ti(result.processes, cmd_column, ti_prov)

        result.summary = _get_host_event_summary(
            self.query_provider,
            self.timespan,
            result.host_entity.HostName,
            result.host_entity.OSFamily,
        )

        self._last_result = result

        if not self.silent:
            self._display_output()

        return self._last_result

    @set_text(docs=_CELL_DOCS, key="show_host_entity")
    def _display_entity(self):
        """Display the host_entity output."""
        if self.check_valid_result_data("host_entity", silent=True):
            nb_print(self._last_result.host_entity)

    @set_text(docs=_CELL_DOCS, key="run")
    def _display_summary(self):
        """Display the summary output."""
        if self.check_valid_result_data("summary", silent=True):
            display(self._last_result.summary)

    @set_text(docs=_CELL_DOCS, key="show_bookmarks")
    def _display_bookmarks(self):
        """Display the bookmarks related to the host."""
        if self.check_valid_result_data("related_bookmarks", silent=True):
            display(self._last_result.related_bookmarks)
        else:
            nb_markdown(f"No Bookmarks related to {self.host_name}")

    @set_text(docs=_CELL_DOCS, key="show_scheduled_tasks")
    def _display_scheduled_tasks(self):
        """Display the scheduled_tasks related to the host."""
        if self.check_valid_result_data("scheduled_tasks", silent=True):
            display(self._last_result.scheduled_tasks)
        else:
            nb_markdown(f"No scheduled tasks related to {self.host_name}")

    @set_text(docs=_CELL_DOCS, key="show_account_actions")
    def _display_account_actions(self):
        """Display the account_actions related to the host."""
        if self.check_valid_result_data("account_actions", silent=True):
            display(self._last_result.account_actions)
        else:
            nb_markdown(f"No account actions related to {self.host_name}")

    @set_text(docs=_CELL_DOCS, key="show_notable_events")
    def _display_notable_events(self):
        """Display the notable_events related to the host."""
        if self.check_valid_result_data("notable_events", silent=True):
            display(self._last_result.notable_events)
        else:
            nb_markdown(f"No notable events related to {self.host_name}")

    @set_text(docs=_CELL_DOCS, key="show_processes")
    def _display_processes(self):
        """Display the processes related to the host."""
        if self.check_valid_result_data("processes", silent=True):
            nb_print(self._last_result.processes)
        else:
            nb_markdown(f"No processes related to {self.host_name}")

    @set_text(docs=_CELL_DOCS, key="show_process_ti")
    def _display_process_ti(self):
        """Display the processes related to the host."""
        if self.check_valid_result_data("process_ti", silent=True):
            nb_print(self._last_result.process_ti)
        else:
            nb_markdown(f"No TI found in process data related to {self.host_name}")

    def _display_output(self):
        """Display all notebooklet sections."""
        self._display_entity()
        self._display_summary()
        self._display_bookmarks()
        self._display_scheduled_tasks()
        self._display_account_actions()
        self._display_notable_events()
        self._display_process_ti()

    def display_process_tree(self):
        """Diplay a process tree from process data."""
        if self.check_valid_result_data("processes", silent=True):
            self._last_result.processes.mp_plot.process_tree()

    def browse_alerts(self) -> nbwidgets.SelectAlert:
        """Return alert browser/viewer."""
        if self.check_valid_result_data("related_alerts", silent=True):
            return browse_alerts(self._last_result)
        return None

    def display_alert_timeline(self):
        """Display the alert timeline."""
        if self.check_valid_result_data("related_alerts", silent=True):
            if len(self._last_result.related_alerts) > 1:
                return _show_alert_timeline(self._last_result.related_alerts)
            print("Cannot plot timeline with 0 or 1 event.")
        return None

    def display_alert_summary(self):
        """Display summarized view of alerts grouped by AlertName."""
        if self.check_valid_result_data("related_alerts", silent=True):
            return (
                self.related_alerts[["AlertName", "TimeGenerated"]]
                .groupby("AlertName")
                .TimeGenerated.agg("count")
            )
        return None


def _process_ti(data, col, ti_prov) -> Optional[pd.DataFrame]:
    extracted_iocs = extract_iocs(data, col, True)
    _, ti_merged_df = get_ti_results(ti_lookup=ti_prov, data=extracted_iocs, col="IoC")
    return ti_merged_df


@lru_cache()
def _get_process_events(qry_prov, timespan, host_name, os_family) -> pd.DataFrame:
    process_events = pd.DataFrame()
    if os_family.name == "Windows":
        nb_data_wait("Process Events")
        process_events = qry_prov.WindowsSecurity.list_host_processes(
            timespan, host_name=host_name, host_op="=~"
        )
    elif os_family.name == "Linux":
        nb_data_wait("Process Events")
        process_events = qry_prov.LinuxSyslog.sysmon_process_events(
            timespan, host_name=f"'{host_name}'"
        )
        if process_events.empty:
            process_events = qry_prov.LinuxSyslog.summarize_events(
                timespan, host_name=f"'{host_name}'"
            )
    return process_events


@lru_cache()
def _get_host_event_summary(qry_prov, timespan, host_name, os_family) -> pd.DataFrame:
    host_events = pd.DataFrame()
    if os_family.name == "Windows":
        nb_data_wait("Events")
        host_events = qry_prov.WindowsSecurity.summarize_events(
            timespan, host_name=host_name, host_op="has"
        )
    elif os_family.name == "Linux":
        nb_data_wait("Events")
        host_events = qry_prov.LinuxSyslog.summarize_events(
            timespan, host_name=f"'{host_name}'"
        )
    return host_events


@lru_cache()
def _get_notable_events(qry_prov, timespan, host_name, os_family) -> pd.DataFrame:
    notable_events = pd.DataFrame()
    if os_family.name == "Windows":
        nb_data_wait("Notable Events")
        notable_events = qry_prov.WindowsSecurity.notable_events(
            timespan, host_name=host_name, host_op="has"
        )
    elif os_family.name == "Linux":
        nb_data_wait("Notable Events")
        notable_events = qry_prov.LinuxSyslog.notable_events(
            timespan, host_name=f"'{host_name}'"
        )
    return notable_events


@lru_cache()
def _get_account_actions(qry_prov, timespan, host_name, os_family) -> pd.DataFrame:
    account_actions = pd.DataFrame()
    if os_family.name == "Windows":
        nb_data_wait("Account Change Events")
        account_actions = qry_prov.WindowsSecurity.account_change_events(
            timespan, host_name=host_name, host_op="=~"
        )
    elif os_family.name == "Linux":
        nb_data_wait("Account Change Events")
        account_actions = qry_prov.LinuxSyslog.user_group_activity(
            timespan, host_name=host_name
        )
    return account_actions


@lru_cache()
def _get_scheduled_tasks(qry_prov, timespan, host_name, os_family) -> pd.DataFrame:
    scheduled_tasks = pd.DataFrame()
    if os_family.name == "Windows":
        nb_data_wait("Scheduled Tasks")
        scheduled_tasks = qry_prov.WindowsSecurity.schdld_tasks_and_services(
            timespan, host_name=host_name, host_op="=~"
        )
    elif os_family.name == "Linux":
        nb_data_wait("Scheduled Tasks")
        scheduled_tasks = qry_prov.LinuxSyslog.cron_activity(
            timespan, host_name=host_name
        )
    return scheduled_tasks


# Get Azure Resource details from API
@lru_cache()
def _azure_api_details(az_cli, host_record):
    try:
        # Get subscription details
        sub_details = az_cli.get_subscription_info(
            host_record.AzureDetails["SubscriptionId"]
        )
        # Get resource details
        resource_details = az_cli.get_resource_details(
            resource_id=host_record.AzureDetails["ResourceId"],
            sub_id=host_record.AzureDetails["SubscriptionId"],
        )
        # Get details of attached disks and network interfaces
        disks = [
            disk.get("name")
            for disk in resource_details.get("properties", {})
            .get("storageProfile", {})
            .get("dataDisks", {})
        ]
        network_ints = [
            net.get("id")
            for net in resource_details.get("properties", {})
            .get("networkProfile", {})
            .get("networkInterfaces")
        ]
        image = (
            str(
                resource_details.get("properties", {})
                .get("storageProfile", {})
                .get("imageReference", {})
                .get("offer", {})
            )
            + " "
            + str(
                resource_details.get("properties", {})
                .get("storageProfile", {})
                .get("imageReference", {})
                .get("sku", {})
            )
        )
        # Extract key details and add host_entity
        resource_details = {
            "Azure Location": resource_details.get("location"),
            "VM Size": (
                resource_details.get("properties", {})
                .get("hardwareProfile", {})
                .get("vmSize")
            ),
            "Image": image,
            "Disks": disks,
            "Admin User": resource_details.get("properties", {})
            .get("osProfile", {})
            .get("adminUsername"),
            "Network Interfaces": network_ints,
            "Tags": str(resource_details.get("tags")),
        }
        return {"resource_details": resource_details, "sub_details": sub_details}
    except CloudError:
        return None


# %%
# Get related alerts
@lru_cache()
def _get_related_alerts(qry_prov, timespan, host_name):
    return qry_prov.SecurityAlert.list_related_alerts(timespan, host_name=host_name)


@set_text(docs=_CELL_DOCS, key="show_alert_timeline")
def _show_alert_timeline(related_alerts):
    if len(related_alerts) > 1:
        return display_timeline(
            data=related_alerts,
            title="Related Alerts",
            source_columns=["AlertName", "TimeGenerated"],
            height=200,
        )
    if len(related_alerts) == 1:
        nb_markdown("A single alert cannot be plotted on a timeline.")
    else:
        nb_markdown("No alerts available to be plotted.")
    return None


@lru_cache()
def _get_related_bookmarks(qry_prov, timespan, host_name):
    nb_data_wait("Bookmarks")
    return qry_prov.AzureSentinel.list_bookmarks_for_entity(
        timespan, entity_id=f"'{host_name}'"
    )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""common test class."""
# from contextlib import redirect_stdout
from contextlib import redirect_stdout
from io import StringIO

import pandas as pd
import pytest
import pytest_check as check
from lxml import etree  # nosec
from markdown import markdown
from msticpy.common.timespan import TimeSpan

from msticnb import data_providers, init, nblts
from msticnb.common import MsticnbDataProviderError
from msticnb.nb.azsent.host.host_summary import HostSummaryResult
from msticnb.read_modules import Notebooklet

from .nb_test import TstNBSummary
from .unit_test_lib import GeoIPLiteMock

# pylint: disable=c-extension-no-member, protected-access


def test_notebooklet_create(monkeypatch):
    """Test method."""
    # Should run because required providers are loaded
    init()
    monkeypatch.setattr(data_providers, "GeoLiteLookup", GeoIPLiteMock)
    data_providers.init(
        query_provider="LocalData", providers=["tilookup", "geolitelookup"]
    )
    for _, nblt in nblts.iter_classes():
        new_nblt = nblt()
        check.is_instance(new_nblt, Notebooklet)
        check.is_none(new_nblt.result)

    # Should throw a warning because of unrecognized provider
    data_providers.init(query_provider="LocalData")
    with pytest.raises(MsticnbDataProviderError) as err:
        for _, nblt in nblts.iter_classes():
            curr_provs = nblt.metadata.req_providers
            bad_provs = [*curr_provs, "bad_provider"]
            try:
                nblt.metadata.req_providers = bad_provs
                new_nblt = nblt()
                check.is_instance(new_nblt, Notebooklet)
                check.is_none(new_nblt.result)
            except MsticnbDataProviderError:
                raise
            finally:
                nblt.metadata.req_providers = curr_provs
    check.is_in("bad_provider", err.value.args[0])
    test_nb = TstNBSummary()
    check.is_not_none(test_nb.get_provider("LocalData"))
    with pytest.raises(MsticnbDataProviderError):
        test_nb.get_provider("otherprovider")


def test_notebooklet_params(monkeypatch):
    """Test supplying timespan param."""
    monkeypatch.setattr(data_providers, "GeoLiteLookup", GeoIPLiteMock)
    data_providers.init(
        query_provider="LocalData", providers=["tilookup", "geolitelookup"]
    )
    test_nb = TstNBSummary()

    tspan = TimeSpan(period="1D")
    test_nb.run(timespan=tspan)
    check.equal(tspan, test_nb.timespan)

    test_nb.run(start=tspan.start, end=tspan.end)
    check.equal(tspan, test_nb.timespan)


def test_notebooklet_options(monkeypatch):
    """Test option logic for notebooklet."""
    monkeypatch.setattr(data_providers, "GeoLiteLookup", GeoIPLiteMock)
    data_providers.init(
        query_provider="LocalData", providers=["tilookup", "geolitelookup"]
    )
    nb_test = TstNBSummary()

    # default options
    nb_res = nb_test.run()
    check.is_not_none(nb_res.default_property)
    check.is_none(nb_res.optional_property)

    # add optional option
    nb_res = nb_test.run(options=["+optional_opt"])
    check.is_not_none(nb_res.default_property)
    check.is_not_none(nb_res.optional_property)

    # remove default option
    nb_res = nb_test.run(options=["-default_opt"])
    check.is_none(nb_res.default_property)
    check.is_none(nb_res.optional_property)

    # specific options
    nb_res = nb_test.run(options=["heartbest", "azure_net"])
    check.is_none(nb_res.default_property)
    check.is_none(nb_res.optional_property)

    # invalid option
    f_stream = StringIO()
    with redirect_stdout(f_stream):
        nb_test.run(options=["invalid_opt"])
    output = str(f_stream.getvalue())
    check.is_in("Invalid options ['invalid_opt']", output)


def test_class_doc():
    """Test class documentation."""
    for _, nblt in nblts.iter_classes():
        html_doc = nblt.get_help()
        check.not_equal(html_doc, "No documentation available.")
        check.greater(len(html_doc), 100)

        md_doc = nblt.get_help(fmt="md")
        html_doc2 = markdown(md_doc)
        check.equal(html_doc, html_doc2)

        _html_parser = etree.HTMLParser(recover=False)
        elem_tree = etree.parse(StringIO(html_doc), _html_parser)
        check.is_not_none(elem_tree)


def test_class_methods():
    """Test method."""
    for _, nblt in nblts.iter_classes():
        check.is_not_none(nblt.description())
        check.is_not_none(nblt.name())
        all_opts = len(nblt.all_options())
        check.greater_equal(all_opts, len(nblt.default_options()))
        check.greater(len(nblt.keywords()), 0)
        check.greater(len(nblt.entity_types()), 0)
        metadata = nblt.get_settings(print_settings=False)
        check.is_not_none(metadata)
        check.is_in("mod_name", metadata)
        check.is_in("default_options", metadata)
        check.is_in("keywords", metadata)


def test_nbresult():
    """Test method."""
    host_result = HostSummaryResult()
    host_result.host_entity = {"host_name": "myhost"}
    host_result.related_alerts = pd.DataFrame()
    host_result.related_bookmarks = pd.DataFrame()
    check.is_in("host_entity:", str(host_result))
    check.is_in("DataFrame:", str(host_result))
    check.is_in("host_entity", host_result.properties)

    html_doc = host_result._repr_html_()
    _html_parser = etree.HTMLParser(recover=False)
    elem_tree = etree.parse(StringIO(html_doc), _html_parser)
    check.is_not_none(elem_tree)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Notebooklet templates module."""
import os
from pathlib import Path
from typing import Optional, Union

from msticpy.common.utility import valid_pyname

from .nb.template import nb_template

_BLACK_IMPORTED = True
try:
    import black
except ImportError:
    _BLACK_IMPORTED = False


DELETE_LINES = [
    '# change the ".." to "...."',
    "from ..._version import VERSION",
    "# Note - when moved to the final location (e.g.",
    "# nb/environ/category/mynotebooklet.py)",
    '# you will need to change the "..." to "...." in these',
    "# imports because the relative path has changed.",
    '# change the "..." to "...."',
]

REPLACE_TEXT = {
    "from ... import nb_metadata": "from msticnb import nb_metadata",
    "from ...common import (": "from msticnb.common import (",
    "from ...notebooklet import": "from msticnb.notebooklet import",
    "__version__ = VERSION": '__version__ = "1.0"',
    '__author__ = "Your name"': '__author__ = "{author}"',
    "TemplateResult": "{nb_name}Result",
    "Template Results.": "{nb_name} Results.",
    "TemplateNB": "{nb_name}",
    "Template Notebooklet class": "{nb_name} Notebooklet class",
}


def create_template(
    nb_name: str = "MyNotebooklet",
    folder: Union[str, Path] = ".",
    author: Optional[str] = None,
    subfolder: bool = False,
    overwrite: bool = False,
):
    """
    Create a notebooklet template.

    Parameters
    ----------
    nb_name : str, optional
        The name of the notebooklet class, by default "MyNotebooklet"
    folder : Union[str, Path], optional
        The target folder for the notebooklet, by default "."
    author : str, optional
        Author name to put in the notebooklet Python module, by default None
    subfolder : bool, optional
        If True create a subfolder for the notebooklet, by default False
    overwrite : bool, optional
        If True overwrite existing files with the same name, by default False.

    """
    author = author or os.environ.get("USER") or os.environ.get("USERNAME") or "Author"
    nb_name = valid_pyname(nb_name)
    folder = Path(folder)
    output_template = _edit_template(nb_name, author)

    target_name = nb_name.casefold()
    # Create folder
    if not Path(folder).is_absolute():
        # If the folder is already the name of the notebooklet,
        folder = Path(".").joinpath(folder).resolve()

    if subfolder:
        folder = folder.joinpath(target_name)
    folder.mkdir(parents=True, exist_ok=True)

    # write files
    print(f"Creating files in {folder}: ")
    py_file = folder.joinpath(f"{target_name}.py")
    if not py_file.is_file() or overwrite:
        py_file.write_text(output_template, encoding="utf-8")
        print(py_file.name, end=", ")
    else:
        print(py_file.name, "not overwritten", end=", ")

    yaml_text = _edit_yaml(nb_name)
    yaml_file = folder.joinpath(f"{target_name}.yaml")
    if not yaml_file.is_file() or overwrite:
        yaml_file.write_text(yaml_text, encoding="utf-8")
        print(yaml_file.name, end=", ")
    else:
        print(yaml_file.name, "not overwritten", end=", ")
    init_file = folder.joinpath("__init__.py")
    if not init_file.is_file():
        init_file.write_text("", encoding="utf-8")
        print(init_file.name, end="")
    else:
        print(init_file.name, "exists - skipping", end="")
    print()


def _edit_template(nb_name: str, author: str) -> str:
    src_template = Path(nb_template.__file__).read_text(encoding="utf-8")

    output_template = "\n".join(
        line for line in src_template.split("\n") if line not in DELETE_LINES
    )

    for src, repl in REPLACE_TEXT.items():
        repl_text = repl.format(nb_name=nb_name, author=author)
        output_template = output_template.replace(src, repl_text)

    if _BLACK_IMPORTED:
        return black.format_str(output_template, mode=black.Mode())
    return output_template


def _edit_yaml(nb_name: str) -> str:
    yaml_text = (
        Path(nb_template.__file__).with_suffix(".yaml").read_text(encoding="utf-8")
    )
    yaml_text = yaml_text.replace("TemplateNB", nb_name).replace(
        "Template YAML for Notebooklet", f"YAML for {nb_name} Notebooklet"
    )
    return yaml_text



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""host_network_summary notebooklet."""
from collections import namedtuple
from functools import lru_cache
from typing import Any, Dict, List, Optional, Set

import pandas as pd
from msticpy.common.timespan import TimeSpan
from msticpy.data import QueryProvider  # pylint: disable=no-name-in-module
from msticpy.datamodel import entities

from ..._version import VERSION
from ...common import MsticnbMissingParameterError, df_has_data, nb_data_wait, nb_print
from ..iptools import convert_to_ip_entities

__version__ = VERSION
__author__ = "Ian Hellen"


@lru_cache()
def get_heartbeat(
    qry_prov: QueryProvider,
    host_name: Optional[str] = None,
    host_ip: Optional[str] = None,
) -> entities.Host:
    """
    Get Heartbeat information for host or IP.

    Parameters
    ----------
    qry_prov : QueryProvider
        Query provider to use for queries
    host_name : str, optional
        Host name, by default None
    host_ip : str, optional
        Host IP Address, by default None

    Returns
    -------
    Host
        Host entity

    """
    host_entity = entities.Host(HostName=host_name)
    if "Heartbeat" in qry_prov.schema_tables:
        nb_data_wait("Heartbeat")
        host_hb_df = None
        if host_name:
            host_hb_df = qry_prov.Network.get_heartbeat_for_host(host_name=host_name)
        elif host_ip:
            host_hb_df = qry_prov.Network.get_heartbeat_for_ip(ip_address=host_ip)
        if df_has_data(host_hb_df):
            host_entity = populate_host_entity(heartbeat_df=host_hb_df)

    return host_entity


# %%
# Get IP Information from Azure Network Topology
# pylint: disable=too-many-branches
@lru_cache()
def get_aznet_topology(
    qry_prov: QueryProvider,
    host_entity: entities.Host,
    host_name: Optional[str] = None,
    host_ip: Optional[str] = None,
):
    """
    Get Azure Network topology information for host or IP address.

    Parameters
    ----------
    qry_prov : QueryProvider
        Query provider to use for queries
    host_entity : Host
        Host entity to populate data with
    host_name : str, optional
        Host name, by default None
    host_ip : str, optional
        Host IP Address, by default None

    """
    if "AzureNetworkAnalytics_CL" not in qry_prov.schema_tables:
        return None
    nb_data_wait("AzureNetworkAnalytics")
    az_net_df = None
    if host_name:
        az_net_df = qry_prov.Network.get_ips_for_host(host_name=host_name)
    elif host_ip:
        az_net_df = qry_prov.Network.host_for_ip(ip_address=host_ip)

    if df_has_data(az_net_df):
        host_entity = populate_host_entity(az_net_df=az_net_df, host_entity=host_entity)
    return host_entity


HostNameVerif = namedtuple("HostNameVerif", "host_name, host_type, host_names")


@lru_cache()  # noqa:MC0001
def verify_host_name(  # noqa: MC0001, C901
    qry_prov: QueryProvider, host_name: str, timespan: TimeSpan = None, **kwargs
) -> HostNameVerif:
    """
    Verify unique hostname by checking Win and Linux logs.

    Parameters
    ----------
    qry_prov : QueryProvider
        Kql query provider
    timespan : TimeSpan
        Time span over which to query
    host_name : str
        The full or partial hostname.

    Returns
    -------
    HostNameVerif
        Tuple[Optional[str], Optional[str], Optional[Dict[str, str]]]
        Named tuple HostNameVerif
        fields: host_name, host_type, host_names
        If unique hostname found, host_name is populated.
        If multiple matching hostnames found, host_names is
        populated and host_name is None.
        host_type is either Windows or Linux.
        If no matching host then all fields are None.

    """
    # Check if a time span is provide as TimeSpan object or start and end parameters
    if timespan is None and ("start" in kwargs and "end" in kwargs):
        start = kwargs["start"]
        end = kwargs["end"]
    elif timespan is not None:
        start = timespan.start
        end = timespan.end
    else:
        raise MsticnbMissingParameterError("timespan")
    host_names: Dict = {}
    # Check for Windows hosts matching host_name
    if "SecurityEvent" in qry_prov.schema_tables:
        sec_event_host = f"""
            SecurityEvent
            | where TimeGenerated between (datetime({start})..datetime({end}))
            | where Computer has "{host_name}"
            | distinct Computer
             """
        nb_data_wait("SecurityEvent")

        win_hosts_df = qry_prov.exec_query(sec_event_host)
        if win_hosts_df is not None and not win_hosts_df.empty:
            for host in win_hosts_df["Computer"].to_list():
                host_names.update({host: "Windows"})

    # Check for Linux hosts matching host_name
    if "Syslog" in qry_prov.schema_tables:
        syslog_host = f"""
            Syslog
            | where TimeGenerated between (datetime({start})..datetime({end}))
            | where Computer has "{host_name}"
            | distinct Computer
            """
        nb_data_wait("Syslog")

        lx_hosts_df = qry_prov.exec_query(syslog_host)

        if lx_hosts_df is not None and not lx_hosts_df.empty:
            for host in lx_hosts_df["Computer"].to_list():
                host_names.update({host: "Linux"})

    # If we have more than one host let the user decide
    if len(host_names.keys()) > 1:
        print(
            f"Multiple matches for '{host_name}'.",
            "Please select a host and re-run.",
            "\n".join(host_names.keys()),
        )
        return HostNameVerif(None, None, host_names)

    if host_names:
        unique_host = next(iter(host_names))
        nb_print(f"Unique host found: {unique_host}")
        return HostNameVerif(unique_host, host_names[unique_host], None)

    nb_print(f"Host not found: {host_name}")
    return HostNameVerif(None, None, None)


# %%
# Populate or create a host entity from Heartbeat and Azure Topology information
def populate_host_entity(
    heartbeat_df: pd.DataFrame = None,
    az_net_df: pd.DataFrame = None,
    vmcomputer_df: pd.DataFrame = None,
    host_entity: entities.Host = None,
    geo_lookup: Any = None,
) -> entities.Host:
    """
    Populate host with IP and other data.

    Parameters
    ----------
    heartbeat_df : pd.DataFrame
        Optional dataframe of heartbeat data for the host
    az_net_df : pd.DataFrame
        Optional dataframe of Azure network data for the host
    vmcomputer_df : pd.DataFrame
        Optional dataframe of VMComputer data for the host
    host_entity : Host
        Host entity in which to populate data. By default,
        a new host entity will be created.
    geo_lookup : Any
        GeoIP Provider to use, if needed.

    Returns
    -------
    Host
        How with details of the IP data collected

    """
    if host_entity is None:
        host_entity = entities.Host()

    ip_entities: List[entities.IpAddress] = []
    ip_unique: Set[str] = set()
    # Extract data from available dataframes
    if df_has_data(heartbeat_df):
        ip_hb = heartbeat_df.iloc[0]  # type: ignore
        ip_entity = _extract_heartbeat(ip_hb, host_entity)
        ip_entities.append(ip_entity)
        ip_unique.add(ip_entity.Address)

    if df_has_data(vmcomputer_df):
        ip_vm = vmcomputer_df.iloc[0]  # type: ignore
        _extract_vmcomputer(ip_vm, host_entity)
        ip_ents = convert_to_ip_entities(
            data=vmcomputer_df, ip_col="Ipv4Addresses", geo_lookup=geo_lookup
        )
        ip_entities.extend(
            ip_ent for ip_ent in ip_ents if ip_ent.Address not in ip_unique
        )
        ip_unique |= {ip_ent.Address for ip_ent in ip_ents}
        ip_ents = convert_to_ip_entities(
            data=vmcomputer_df, ip_col="Ipv6Addresses", geo_lookup=geo_lookup
        )
        ip_entities.extend(
            ip_ent for ip_ent in ip_ents if ip_ent.Address not in ip_unique
        )
        ip_unique |= {ip_ent.Address for ip_ent in ip_ents}

    # If Azure network data present add this to host record
    if df_has_data(az_net_df):
        if not host_entity.HostName:
            host_entity.HostName = az_net_df.iloc[0].Computer  # type: ignore
        ip_priv = convert_to_ip_entities(
            data=az_net_df, ip_col="PrivateIPAddresses", geo_lookup=geo_lookup
        )
        ip_pub = convert_to_ip_entities(
            data=az_net_df, ip_col="PublicIPAddresses", geo_lookup=geo_lookup
        )
        host_entity["PrivateIpAddresses"] = []
        host_entity["PrivateIpAddresses"].extend(
            ip_ent for ip_ent in ip_priv if ip_ent.Address not in ip_unique
        )
        host_entity["PublicIpAddresses"] = []
        host_entity["PublicIpAddresses"].extend(
            ip_ent for ip_ent in ip_pub if ip_ent.Address not in ip_unique
        )
        ip_entities.extend(host_entity["PrivateIpAddresses"])
        ip_entities.extend(host_entity["PublicIpAddresses"])

    host_entity["IpAddresses"] = ip_entities
    if not hasattr(host_entity, "IpAddress") and len(ip_entities) == 1:
        host_entity["IpAddress"] = ip_entities[0]

    return host_entity


def _extract_heartbeat(ip_hb, host_entity):
    if not host_entity.HostName:
        host_entity.HostName = ip_hb["Computer"]  # type: ignore
    host_entity.SourceComputerId = ip_hb["SourceComputerId"]  # type: ignore
    host_entity.OSFamily = (
        entities.OSFamily.Windows
        if ip_hb["OSType"] == "Windows"
        else entities.OSFamily.Linux
    )
    host_entity.OSName = ip_hb["OSName"]  # type: ignore
    host_entity.OSVMajorVersion = ip_hb["OSMajorVersion"]  # type: ignore
    host_entity.OSVMinorVersion = ip_hb["OSMinorVersion"]  # type: ignore
    host_entity.Environment = ip_hb["ComputerEnvironment"]  # type: ignore
    host_entity.AgentId = ip_hb["SourceComputerId"]
    host_entity.OmsSolutions = [  # type: ignore
        sol.strip() for sol in ip_hb["Solutions"].split(",")
    ]
    host_entity.VMUUID = ip_hb["VMUUID"]  # type: ignore
    if host_entity.Environment == "Azure":
        host_entity.AzureDetails = {  # type: ignore
            "SubscriptionId": ip_hb["SubscriptionId"],
            "ResourceProvider": ip_hb["ResourceProvider"],
            "ResourceType": ip_hb["ResourceType"],
            "ResourceGroup": ip_hb["ResourceGroup"],
            "ResourceId": ip_hb["ResourceId"],
        }

    # Populate IP data
    ip_entity = entities.IpAddress(Address=ip_hb["ComputerIP"])
    geoloc_entity = entities.GeoLocation()  # type: ignore
    geoloc_entity.CountryName = ip_hb["RemoteIPCountry"]  # type: ignore
    geoloc_entity.Longitude = ip_hb["RemoteIPLongitude"]  # type: ignore
    geoloc_entity.Latitude = ip_hb["RemoteIPLatitude"]  # type: ignore
    ip_entity.Location = geoloc_entity  # type: ignore
    host_entity.IpAddress = ip_entity  # type: ignore
    return ip_entity


def _extract_vmcomputer(ip_vm, host_entity):
    if not host_entity.HostName:
        host_entity.HostName = ip_vm["Computer"]  # type: ignore
    host_entity.OSFamily = (
        entities.OSFamily.Windows
        if ip_vm["OperatingSystemFamily"].casefold() == "windows"
        else entities.OSFamily.Linux
    )
    host_entity.OSName = ip_vm["OperatingSystemFullName"]  # type: ignore
    host_entity.Environment = "Azure"  # type: ignore
    host_entity.AgentId = ip_vm["AgentId"]
    host_entity.VMUUID = ip_vm["AzureVmId"]  # type: ignore
    if host_entity.Environment == "Azure":
        host_entity.AzureDetails = {  # type: ignore
            "SubscriptionId": ip_vm["AzureSubscriptionId"],
            "ResourceProvider": ip_vm["HostingProvider"],
            "ResourceType": ip_vm["VirtualMachineType"],
            "ResourceGroup": ip_vm["AzureResourceGroup"],
            "ResourceId": ip_vm["_ResourceId"],
        }
